"""Click definitions for various shared options and arguments."""


import functools
import os

import click
from galaxy.tool_util.deps import docker_util
from galaxy.tool_util.verify.interactor import DEFAULT_TOOL_TEST_WAIT

from .config import planemo_option


def force_option(what="files"):
    """Annotate click command as consume the -f/--force option."""
    return planemo_option(
        "-f",
        "--force",
        is_flag=True,
        help="Overwrite existing %s if present." % what,
    )


def skip_venv_option():
    """Annotate click command as consume the --skip_venv option."""
    return planemo_option(
        "--skip_venv",
        is_flag=True,
        help=(
            "Do not create or source a virtualenv environment for Galaxy, "
            "this should be used to preserve an externally configured "
            "virtual environment or conda environment."
        ),
    )


def skip_client_build_option():
    """Annotate click command as consume the --skip_client_build option."""
    return planemo_option(
        "--skip_client_build", is_flag=True, default=False, help=("Do not build Galaxy client when serving Galaxy.")
    )


def run_engine_option():
    """Annotate click command as consume the --engine option."""
    return planemo_option(
        "--engine",
        type=click.Choice(["galaxy", "docker_galaxy", "cwltool", "toil", "external_galaxy"]),
        default=None,
        use_global_config=True,
        help=(
            "Select an engine to run or test artifacts such as tools "
            "and workflows. Defaults to a local Galaxy, but running Galaxy within "
            "a Docker container or the CWL reference implementation 'cwltool' and "
            "'toil' be selected."
        ),
    )


def non_strict_cwl_option():
    """Annotate click command as consume the --non_strict_cwl option."""
    return planemo_option(
        "--non_strict_cwl",
        default=False,
        is_flag=True,
        help="Disable strict validation of CWL.",
    )


def serve_engine_option():
    """Annotate click command as consume the --engine option.

    This variant of the engine command is restricted to engines that can serve Galaxy
    servers.
    """
    return planemo_option(
        "--engine",
        type=click.Choice(["galaxy", "docker_galaxy", "external_galaxy"]),
        default="galaxy",
        use_global_config=True,
        use_env_var=True,
        help=(
            "Select an engine to serve artifacts such as tools "
            "and workflows. Defaults to a local Galaxy, but running Galaxy within "
            "a Docker container."
        ),
    )


def ignore_dependency_problems_option():
    """Annotate click command as consume the --ignore_dependency_problems option."""
    return planemo_option(
        "--ignore_dependency_problems",
        is_flag=True,
        default=False,
        use_global_config=True,
        help=(
            "When installing shed repositories for workflows, ignore dependency issues. "
            "These likely indicate a problem but in some cases may not prevent a workflow "
            "from successfully executing."
        ),
    )


def cwltool_no_container_option():
    """Annotate click command as consume the --no_container option.

    This option is for the CWL CLI runner interface.
    """
    return planemo_option(
        "--no-container",
        "--no_container",
        is_flag=True,
        default=False,
        use_global_config=True,
        help=("If cwltool engine is used, disable Docker container usage."),
    )


def test_data_option():
    """Annotate click command as consume the --test_data option."""
    return planemo_option(
        "--test_data",
        type=click.Path(exists=True, file_okay=False, resolve_path=True),
        help="test-data directory to for specified tool(s).",
    )


def extra_tools_option():
    return planemo_option(
        "--extra_tools",
        type=click.Path(exists=True, file_okay=True, dir_okay=True, resolve_path=True),
        multiple=True,
        help=(
            "Extra tool sources to include in Galaxy's tool panel (file or "
            "directory). These will not be linted/tested/etc... but they "
            "will be available to workflows and for interactive use."
        ),
    )


def tool_data_table_option():
    return planemo_option(
        "--tool_data_table",
        type=click.Path(exists=True, file_okay=True, resolve_path=True),
        help="tool_data_table_conf.xml file to for specified tool(s).",
        multiple=True,
    )


def galaxy_email_option():
    return planemo_option(
        "--galaxy_email",
        type=str,
        default="planemo@galaxyproject.org",
        use_global_config=True,
        use_env_var=True,
        help="E-mail address to use when launching single-user Galaxy server.",
    )


def galaxy_python_version():
    return planemo_option(
        "--galaxy_python_version",
        use_global_config=True,
        default=None,
        type=click.Choice(["3", "3.7", "3.8", "3.9"]),
        help="Python version to start Galaxy under",
    )


def galaxy_root_option():
    return planemo_option(
        "--galaxy_root",
        use_global_config=True,
        extra_global_config_vars=["galaxy_root"],
        use_env_var=True,
        type=click.Path(file_okay=False, dir_okay=True, resolve_path=True),
        help="Root of development galaxy directory to execute command with.",
    )


def galaxy_cwl_root_option():
    return planemo_option(
        "--cwl_galaxy_root",
        use_global_config=True,
        extra_global_config_vars=["cwl_galaxy_root"],
        use_env_var=True,
        type=click.Path(exists=True, file_okay=False, resolve_path=True),
        help=(
            "Root of development galaxy directory to execute command with"
            " (must be branch of Galaxy with CWL support, this option"
            " is experimental and will be replaced with --galaxy_root when"
            " and if CWL support is merged into Galaxy."
        ),
    )


def galaxy_port_option():
    return planemo_option(
        "--port",
        type=int,
        default="9090",
        use_global_config=True,
        help="Port to serve Galaxy on (default is 9090).",
    )


def galaxy_host_option():
    return planemo_option(
        "--host",
        type=str,
        default="127.0.0.1",
        use_global_config=True,
        help=(
            "Host to bind Galaxy to. Default is 127.0.0.1 that is "
            "restricted to localhost connections for security reasons "
            "set to 0.0.0.0 to bind Galaxy to all ports including "
            "potentially publicly accessible ones."
        ),
    )


def dependency_resolvers_option():
    return planemo_option(
        "--dependency_resolvers_config_file",
        type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
        use_global_config=True,
        help="Dependency resolver configuration for Galaxy to target.",
    )


def enable_cwl_option():
    return planemo_option(
        "--cwl",
        is_flag=True,
        help=(
            "Configure Galaxy for use with CWL tool."
            " (this option is experimental and will be replaced when"
            " and if CWL support is merged into Galaxy)."
        ),
    )


def build_cwl_option():
    return planemo_option(
        "--cwl",
        is_flag=True,
        help="Build a CWL tool instead of a Galaxy tool.",
    )


def run_history_tags_option():
    return planemo_option(
        "tags",
        "--tags",
        type=str,
        default=None,
        help="Comma-separated list of tags to add to the created history.",
    )


def run_output_directory_option():
    return planemo_option(
        "output_directory",
        "--output_directory",
        "--outdir",
        type=click.Path(
            file_okay=False,
            dir_okay=True,
            resolve_path=True,
        ),
        default=None,
        help=("Where to store outputs of a 'run' task."),
    )


def run_output_json_option():
    return planemo_option(
        "output_json",
        "--output_json",
        type=click.Path(
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
        ),
        default=None,
        help=("Where to store JSON dictionary describing outputs of " "a 'run' task."),
    )


def run_download_outputs_option():
    return planemo_option(
        "--download_outputs/--no_download_outputs",
        is_flag=True,
        default=False,
        help=(
            "After tool or workflow runs are complete, download "
            "the output files to the location specified by --output_directory. "
        ),
    )


def publish_dockstore_option():
    return planemo_option(
        "--publish/--no_publish",
        is_flag=True,
        default=True,
        help="Set publish attribute to true in .dockstore.yml file",
    )


def no_dependency_resolution():
    return planemo_option(
        "--no_dependency_resolution",
        is_flag=True,
        help="Configure Galaxy with no dependency resolvers.",
    )


def brew_dependency_resolution():
    return planemo_option(
        "--brew_dependency_resolution",
        is_flag=True,
        help="Configure Galaxy to use plain brew dependency resolution.",
    )


def conda_dependency_resolution():
    return planemo_option(
        "--conda_dependency_resolution",
        is_flag=True,
        help="Configure Galaxy to use only conda for dependency resolution.",
    )


def shed_dependency_resolution():
    return planemo_option(
        "--shed_dependency_resolution",
        is_flag=True,
        help=("Configure Galaxy to use brewed Tool Shed dependency" " resolution."),
    )


def file_path_option():
    return planemo_option(
        "--file_path",
        type=click.Path(file_okay=False, dir_okay=True, resolve_path=True),
        help="Location for files created by Galaxy (e.g. database/files).",
        default=None,
        use_global_config=True,
    )


def database_connection_option():
    return planemo_option(
        "--database_connection",
        type=str,
        help="Database connection string to use for Galaxy.",
        default=None,
        use_global_config=True,
    )


def shed_tools_conf_option():
    return planemo_option(
        "--shed_tool_conf",
        type=str,
        help="Location of shed tools conf file for Galaxy.",
        default=None,
        use_global_config=True,
    )


def shed_tools_directory_option():
    return planemo_option(
        "--shed_tool_path",
        type=str,
        help="Location of shed tools directory for Galaxy.",
        default=None,
        use_global_config=True,
    )


def tool_dependency_dir_option():
    return planemo_option(
        "--tool_dependency_dir",
        type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
        default=None,
        use_global_config=True,
        help="Tool dependency dir for Galaxy to target.",
    )


def job_config_option():
    return planemo_option(
        "--job_config_file",
        type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
        help="Job configuration file for Galaxy to target.",
        default=None,
        use_global_config=True,
    )


def tool_data_path_option():
    return planemo_option(
        "--tool_data_path",
        type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
        help="Directory where data used by tools is located. Required if tests are run in docker and should make use of external reference data.",
        default=None,
        use_global_config=True,
    )


def mulled_containers_option():
    return planemo_option(
        "mulled_containers",
        "--mulled_containers",
        "--biocontainers",
        is_flag=True,
        help="Test tools against mulled containers (forces --docker). Disables conda resolution unless any conda option has been set explicitly.",
    )


def galaxy_startup_timeout_option():
    return planemo_option(
        "--galaxy_startup_timeout",
        type=click.IntRange(1),
        default=900,
        help="Wait for galaxy to start before assuming Galaxy did not start.",
    )


def install_galaxy_option():
    return planemo_option(
        "--install_galaxy", is_flag=True, help="Download and configure a disposable copy of Galaxy from github."
    )


def docker_galaxy_image_option():
    return planemo_option(
        "--docker_galaxy_image",
        default="quay.io/bgruening/galaxy",
        use_global_config=True,
        help=(
            "Docker image identifier for docker-galaxy-flavor used if "
            "engine type is specified as ``docker-galaxy``. Defaults to "
            "quay.io/bgruening/galaxy."
        ),
    )


def docker_extra_volume_option():
    arg_type = click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        resolve_path=True,
    )
    return planemo_option(
        "--docker_extra_volume",
        type=arg_type,
        default=None,
        use_global_config=True,
        help=("Extra path to mount if --engine docker."),
    )


def galaxy_url_option():
    return planemo_option(
        "--galaxy_url",
        use_global_config=True,
        extra_global_config_vars=["galaxy_url"],
        use_env_var=True,
        type=str,
        help="Remote Galaxy URL to use with external Galaxy engine.",
    )


def galaxy_admin_key_option():
    return planemo_option(
        "--galaxy_admin_key",
        use_global_config=True,
        extra_global_config_vars=["admin_key"],
        use_env_var=True,
        type=str,
        help="Admin key to use with external Galaxy engine.",
    )


def galaxy_user_key_option():
    return planemo_option(
        "--galaxy_user_key",
        use_global_config=True,
        extra_global_config_vars=["admin_key"],
        use_env_var=True,
        type=str,
        help="User key to use with external Galaxy engine.",
    )


def history_name():
    return planemo_option(
        "--history_name",
        type=str,
        help="Name to give a Galaxy history, if one is created.",
    )


def no_cache_galaxy_option():
    return planemo_option(
        "--no_cache_galaxy",
        is_flag=True,
        help=(
            "Skip caching of Galaxy source and dependencies obtained with "
            "--install_galaxy. Not caching this results in faster "
            "downloads (no git) - so is better on throw away instances such "
            "with TravisCI. "
        ),
    )


def galaxy_branch_option():
    return planemo_option(
        "--galaxy_branch",
        default=None,
        use_global_config=True,
        use_env_var=True,
        help=("Branch of Galaxy to target (defaults to master) if a Galaxy " "root isn't specified."),
    )


def galaxy_source_option():
    return planemo_option(
        "--galaxy_source",
        default=None,
        use_global_config=True,
        help=(
            "Git source of Galaxy to target (defaults to the official "
            "galaxyproject github source if a Galaxy root isn't "
            "specified."
        ),
    )


def skip_install_option():
    return planemo_option(
        "--skip_install", is_flag=True, help="Skip installation - only source requirements already available."
    )


def brew_option():
    return planemo_option(
        "--brew",
        use_global_config=True,
        type=click.Path(exists=True, file_okay=True, dir_okay=False),
        help="Homebrew 'brew' executable to use.",
    )


def conda_prefix_option():
    return planemo_option(
        "--conda_prefix",
        use_global_config=True,
        use_env_var=True,
        type=click.Path(file_okay=False, dir_okay=True),
        help="Conda prefix to use for conda dependency commands.",
    )


def conda_exec_option():
    return planemo_option(
        "--conda_exec",
        use_global_config=True,
        type=click.Path(exists=True, file_okay=True, dir_okay=False),
        help="Location of conda executable.",
    )


def conda_use_local_option():
    return planemo_option(
        "--conda_use_local", is_flag=True, help="Use locally built packages while building Conda environments."
    )


def conda_ensure_channels_option():
    return planemo_option(
        "conda_ensure_channels",
        "--conda_channels",
        "--conda_ensure_channels",
        type=str,
        use_global_config=True,
        use_env_var=True,
        help=("Ensure conda is configured with specified comma separated " "list of channels."),
        default="conda-forge,bioconda,defaults",
    )


def conda_auto_install_option():
    return planemo_option(
        "--conda_auto_install/--no_conda_auto_install",
        is_flag=True,
        default=True,
        help=("Conda dependency resolution for Galaxy will attempt to install " "requested but missing packages."),
    )


def conda_auto_init_option():
    return planemo_option(
        "--conda_auto_init/--no_conda_auto_init",
        is_flag=True,
        default=True,
        help=(
            "Conda dependency resolution for Galaxy will auto install "
            "conda itself using miniconda if not availabe on conda_prefix."
        ),
    )


def conda_global_option():
    return planemo_option(
        "--global",
        is_flag=True,
        default=False,
        help=(
            "Install Conda dependencies globally instead of in requirement specific "
            "environments packaged for tools. If the Conda bin directory is on your "
            "PATH, tools may still use binaries but this is more designed for "
            "interactive testing and debugging."
        ),
    )


def simultaneous_upload_option():
    return planemo_option(
        "--simultaneous_uploads/--no_simultaneous_uploads",
        is_flag=True,
        default=False,
        help=(
            "When uploading files to Galaxy for tool or workflow tests or runs, "
            "upload multiple files simultaneously without waiting for the previous "
            "file upload to complete."
        ),
    )


def check_uploads_ok_option():
    return planemo_option(
        "--check_uploads_ok/--no_check_uploads_ok",
        is_flag=True,
        default=True,
        help=(
            "When uploading files to Galaxy for tool or workflow tests or runs, "
            "check that the history is in an 'ok' state before beginning tool "
            "or workflow execution."
        ),
    )


def required_tool_arg(allow_uris=False):
    """Decorate click method as requiring the path to a single tool."""
    arg_type_class = click.Path if not allow_uris else UriLike
    arg_type = arg_type_class(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    )
    if allow_uris:
        name = "uri"
        metavar = "TOOL_URI"
    else:
        name = "path"
        metavar = "TOOL_PATH"
    return click.argument(name, metavar=metavar, type=arg_type)


def paste_test_data_paths_option():
    return planemo_option(
        "--paste_test_data_paths/--no_paste_test_data_paths",
        is_flag=True,
        default=None,
        help=(
            "By default Planemo will use or not use Galaxy's path paste option to load "
            "test data into a history based on the engine type it is targeting. This can "
            "override the logic to explicitly enable or disable path pasting."
        ),
    )


def shed_install_option():
    return planemo_option(
        "--shed_install/--no_shed_install",
        is_flag=True,
        default=True,
        help=(
            "By default Planemo will attempt to install repositories needed for workflow "
            "testing. This may not be appropriate for production servers and so this can "
            "disabled by calling planemo with --no_shed_install."
        ),
    )


def install_tool_dependencies_option():
    return planemo_option(
        "--install_tool_dependencies/--no_install_tool_dependencies",
        is_flag=True,
        default=False,
        help=("Turn on installation of tool dependencies using classic toolshed packages."),
    )


def install_resolver_dependencies_option():
    return planemo_option(
        "--install_resolver_dependencies/--no_install_resolver_dependencies",
        is_flag=True,
        default=True,
        help=("Skip installing tool dependencies through resolver (e.g. conda)."),
    )


def install_repository_dependencies_option():
    return planemo_option(
        "--install_repository_dependencies/--no_install_repository_dependencies",
        is_flag=True,
        default=True,
        help=("Skip installing the repository dependencies."),
    )


def single_user_mode_option():
    return planemo_option(
        "galaxy_single_user",
        "--galaxy_single_user/--no_galaxy_single_user",
        is_flag=True,
        default=True,
        help=(
            "By default Planemo will configure Galaxy to run in single-user mode where there "
            "is just one user and this user is automatically logged it. Use --no_galaxy_single_user "
            "to prevent Galaxy from running this way."
        ),
    )


def required_workflow_arg():
    return click.argument(
        "workflow_identifier",
        metavar="WORKFLOW_PATH_OR_ID",
        type=str,
    )


def split_job_and_test():
    return click.option(
        "--split_test/--no_split_test", default=False, help="Write workflow job and test definitions to separate files."
    )


def from_invocation():
    return planemo_option(
        "--from_invocation/--from_uri",
        is_flag=True,
        default=False,
        help="Build a workflow test or job description from an invocation ID run on an external Galaxy."
        "A Galaxy URL and API key must also be specified. This allows test data to be downloaded"
        "and inputs and parameters defined automatically. Alternatively, the default is to build the"
        "descriptions from a provided workflow URI.",
    )


def required_job_arg():
    """Decorate click method as requiring the path to a single tool."""
    arg_type = click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=False,
    )
    return click.argument("job_path", metavar="JOB_PATH", type=arg_type)


def required_runnable_arg():
    return click.argument(
        "runnable_identifier",
        metavar="RUNNABLE_PATH_OR_ID",
        type=str,
    )


def required_new_job_arg():
    arg_type = click.Path()
    return click.argument("new_job_path", metavar="NEW_JOB_PATH", type=arg_type)


def _optional_tools_default(ctx, param, value):
    if param.name in ["paths", "uris"] and len(value) == 0:
        return [os.path.abspath(os.getcwd())]
    else:
        return value


def optional_tools_or_packages_arg(multiple=False):
    """Decorate click method as optionally taking in the path to a tool
    or directory of tools or a Conda package. If no such argument is given
    the current working directory will be treated as a directory of tools.
    """
    name = "paths" if multiple else "path"
    nargs = -1 if multiple else 1
    return click.argument(
        name,
        metavar="TARGET",
        nargs=nargs,
    )


class UriLike(click.Path):
    def convert(self, value, param, ctx):
        if "://" in value:
            return value
        else:
            return super().convert(value, param, ctx)


def optional_tools_arg(multiple=False, allow_uris=False):
    """Decorate click method as optionally taking in the path to a tool
    or directory of tools. If no such argument is given the current working
    directory will be treated as a directory of tools.
    """
    arg_type_class = click.Path if not allow_uris else UriLike
    arg_type = arg_type_class(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        resolve_path=True,
    )
    if allow_uris:
        name = "uris" if multiple else "uri"
    else:
        name = "paths" if multiple else "path"
    nargs = -1 if multiple else 1
    return click.argument(
        name,
        metavar="TOOL_PATH",
        type=arg_type,
        nargs=nargs,
        callback=_optional_tools_default,
    )


class ProjectOrRepositry(click.Path):
    def __init__(self, **kwds):
        super().__init__(**kwds)

    def convert(self, value, param, ctx):
        if value and value.startswith("git:") or value.startswith("git+"):
            return value
        else:
            return super().convert(value, param, ctx)


def shed_project_arg(multiple=True):
    arg_type = ProjectOrRepositry(
        exists=True,
        file_okay=False,
        dir_okay=True,
        resolve_path=True,
    )
    name = "paths" if multiple else "path"
    nargs = -1 if multiple else 1
    return click.argument(
        name,
        metavar="PROJECT",
        type=arg_type,
        nargs=nargs,
        callback=_optional_tools_default,
    )


def recipe_arg(multiple=True):
    name = "paths" if multiple else "path"
    nargs = -1 if multiple else 1
    return click.argument(
        name,
        metavar="RECIPE_DIR",
        type=click.Path(
            exists=True,
            file_okay=True,
            dir_okay=True,
            resolve_path=True,
        ),
        nargs=nargs,
        callback=_optional_tools_default,
    )


def optional_project_arg(exists=True, default="."):
    arg_type = click.Path(
        exists=exists,
        file_okay=False,
        dir_okay=True,
        writable=True,
        resolve_path=True,
    )
    return click.argument("path", metavar="PROJECT", default=default, type=arg_type)


def no_cleanup_option():
    return planemo_option("--no_cleanup", is_flag=True, help=("Do not cleanup temp files created for and by Galaxy."))


def docker_enable_option():
    return planemo_option("--docker/--no_docker", default=False, help=("Run Galaxy tools in Docker if enabled."))


def docker_cmd_option():
    return planemo_option(
        "--docker_cmd",
        default=docker_util.DEFAULT_DOCKER_COMMAND,
        help="Command used to launch docker (defaults to docker).",
    )


def docker_sudo_option():
    return planemo_option("--docker_sudo/--no_docker_sudo", is_flag=True, help="Flag to use sudo when running docker.")


def docker_sudo_cmd_option():
    return planemo_option(
        "--docker_sudo_cmd",
        help="sudo command to use when --docker_sudo is enabled " + "(defaults to sudo).",
        default=docker_util.DEFAULT_SUDO_COMMAND,
        use_global_config=True,
    )


def docker_host_option():
    return planemo_option(
        "--docker_host",
        help="Docker host to target when executing docker commands " + "(defaults to localhost).",
        use_global_config=True,
        default=docker_util.DEFAULT_HOST,
    )


def docker_config_options():
    return _compose(
        docker_cmd_option(),
        docker_sudo_option(),
        docker_host_option(),
        docker_sudo_cmd_option(),
    )


def galaxy_docker_options():
    return _compose(
        docker_enable_option(),
        docker_config_options(),
    )


def shed_owner_option():
    return planemo_option("--owner", help="Tool Shed repository owner (username).")


def shed_name_option():
    return planemo_option("--name", help="Tool Shed repository name (defaults to the inferred " "tool directory name).")


def validate_shed_target_callback(ctx, param, value):
    if value is None:
        ctx.fail("default_shed_target set to None, must specify a value for " "--shed_target to run this command.")
    return value


def shed_target_option():
    return planemo_option(
        "-t",
        "--shed_target",
        help="Tool Shed to target (this can be 'toolshed', 'testtoolshed', "
        "'local' (alias for http://localhost:9009/), an arbitrary url "
        "or mappings defined ~/.planemo.yml.",
        default=None,
        use_global_config=True,
        callback=validate_shed_target_callback,
    )


def shed_key_option():
    return planemo_option(
        "--shed_key",
        help=(
            "API key for Tool Shed access. An API key is required unless "
            "e-mail and password is specified. This key can be specified "
            "with either --shed_key or --shed_key_from_env."
        ),
    )


def shed_key_from_env_option():
    return planemo_option("--shed_key_from_env", help="Environment variable to read API key for Tool Shed access from.")


def shed_email_option():
    return planemo_option("--shed_email", help="E-mail for Tool Shed auth (required unless shed_key is " "specified).")


def shed_password_option():
    return planemo_option(
        "--shed_password", help="Password for Tool Shed auth (required unless shed_key is " "specified)."
    )


def shed_skip_upload():
    return planemo_option(
        "--skip_upload", is_flag=True, help=("Skip upload contents as part of operation, only update " "metadata.")
    )


def shed_skip_metadata():
    return planemo_option(
        "--skip_metadata",
        is_flag=True,
        help=("Skip metadata update as part of operation, only upload " "new contents."),
    )


def shed_message_option():
    return planemo_option("-m", "--message", help="Commit message for tool shed upload.")


def shed_force_create_option():
    return planemo_option(
        "--force_repository_creation",
        help=(
            "If a repository cannot be found for the specified user/repo "
            "name pair, then automatically create the repository in the "
            "toolshed."
        ),
        is_flag=True,
        default=False,
    )


def shed_check_diff_option():
    return planemo_option(
        "--check_diff",
        is_flag=True,
        help=(
            "Skip uploading if the shed_diff detects there would be no "
            "'difference' (only attributes populated by the shed would "
            "be updated.)"
        ),
    )


def shed_upload_options():
    return _compose(
        shed_message_option(),
        shed_force_create_option(),
        shed_check_diff_option(),
    )


def shed_realization_options():
    return _compose(
        shed_project_arg(multiple=True),
        recursive_shed_option(),
        shed_fail_fast_option(),
    )


def shed_repo_options():
    return _compose(
        shed_owner_option(),
        shed_name_option(),
    )


def shed_publish_options():
    """Common options for commands that require publishing to a
    a shed.
    """
    return _compose(
        shed_realization_options(),
        shed_repo_options(),
        shed_target_options(),
    )


def shed_read_options():
    """Common options that require read access to mapped repositories
    in a shed.
    """
    return _compose(
        shed_realization_options(),
        shed_repo_options(),
        shed_target_options(),
    )


def shed_target_options():
    """Common options for commands that require read-only
    interactions with a shed.
    """
    return _compose(
        shed_email_option(),
        shed_key_option(),
        shed_key_from_env_option(),
        shed_password_option(),
        shed_target_option(),
    )


def conda_target_options(include_local=True):
    return _compose(
        conda_prefix_option(),
        conda_exec_option(),
        conda_ensure_channels_option(),
        conda_use_local_option(),
    )


def github_namespace():
    return planemo_option(
        "--namespace",
        use_env_var=True,
        help=(
            "Organization or username under which to create or update workflow repository. "
            "Must be a valid github username or organization"
        ),
        default="iwc-workflows",
    )


def dry_run():
    return planemo_option(
        "--dry_run",
        help="Don't execute action, show preview of action.",
        is_flag=True,
        default=False,
    )


def github_branch():
    return planemo_option(
        "--github_branch",
        help="GitHub branch to use for the action. Default is main.",
        default="main",
    )


def galaxy_run_options():
    return _compose(
        galaxy_target_options(),
        galaxy_port_option(),
        galaxy_host_option(),
    )


def galaxy_config_options():
    return _compose(
        test_data_option(),
        tool_data_table_option(),
        dependency_resolvers_option(),
        brew_dependency_resolution(),
        shed_dependency_resolution(),
        no_dependency_resolution(),
        conda_target_options(),
        conda_dependency_resolution(),
        conda_auto_install_option(),
        conda_auto_init_option(),
        simultaneous_upload_option(),
        check_uploads_ok_option(),
        # Profile options...
        profile_option(),
        profile_database_options(),
        file_path_option(),
        database_connection_option(),
        shed_tools_conf_option(),
        shed_tools_directory_option(),
        single_user_mode_option(),
    )


def galaxy_target_options():
    return _compose(
        galaxy_root_option(),
        galaxy_python_version(),
        extra_tools_option(),
        install_galaxy_option(),
        galaxy_branch_option(),
        galaxy_source_option(),
        skip_venv_option(),
        no_cache_galaxy_option(),
        no_cleanup_option(),
        galaxy_email_option(),
        galaxy_docker_options(),
        mulled_containers_option(),
        galaxy_startup_timeout_option(),
        # Profile options...
        job_config_option(),
        tool_dependency_dir_option(),
        tool_data_path_option(),
    )


def pid_file_option():
    pid_file_type = click.Path(
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    )
    return planemo_option(
        "--pid_file", type=pid_file_type, default=None, help="Location of pid file is executed with --daemon."
    )


def daemon_option():
    return planemo_option("--daemon", is_flag=True, help="Serve Galaxy process as a daemon.")


def profile_option(required=False):
    return planemo_option(
        "--profile",
        type=click.STRING,
        required=required,
        default=None,
        help=("Name of profile (created with the profile_create command) to use " "with this command."),
    )


def alias_option(required=False):
    return planemo_option("--alias", type=click.STRING, required=required, default=None, help=("Name of an alias."))


def galaxy_serve_options():
    return _compose(
        galaxy_run_options(),
        serve_engine_option(),
        non_strict_cwl_option(),
        docker_galaxy_image_option(),
        docker_extra_volume_option(),
        galaxy_config_options(),
        daemon_option(),
        pid_file_option(),
        ignore_dependency_problems_option(),
        skip_client_build_option(),
        shed_install_option(),
    )


def training_topic_name_option():
    return planemo_option(
        "--topic_name",
        required=True,
        help="Name (directory name) of the topic to create or in which " "a tutorial should be created or updates",
    )


def training_topic_option():
    return _compose(
        training_topic_name_option(),
        planemo_option("--topic_title", default="Title of the topic", help="Title of the topic to create"),
        planemo_option("--topic_summary", default="Summary of the topic", help="Summary of the topic"),
        planemo_option(
            "--topic_target",
            type=click.Choice(["use", "admin-dev", "instructors"]),
            default="use",
            help="Target audience for the topic",
        ),
    )


def training_tutorial_name_option():
    return planemo_option("--tutorial_name", help="Name (directory name) of the tutorial to create or to modify")


def training_tutorial_name_req_option():
    return planemo_option("--tutorial_name", required=True, help="Name (directory name) of the tutorial to modify")


def training_datatype_option():
    return planemo_option(
        "--datatypes",
        type=click.Path(file_okay=True, resolve_path=True),
        help="YAML file with the correspondance between Zenodo extension and Galaxy datatypes",
        default="shared/datatypes.yaml",
    )


def training_zenodo_option():
    return planemo_option("--zenodo_link", help="Zenodo URL with the input data")


def training_tutorial_worflow_option():
    return _compose(
        planemo_option(
            "--workflow",
            type=click.Path(file_okay=True, resolve_path=True),
            help="Workflow of the tutorial (locally)",
            default=None,
        ),
        planemo_option("--galaxy_url", help="URL of a Galaxy instance with the workflow"),
        planemo_option("--galaxy_api_key", help="API key on the Galaxy instance with the workflow"),
        planemo_option("--workflow_id", help="ID of the workflow on the Galaxy instance"),
    )


def training_tutorial_option():
    return _compose(
        training_tutorial_name_option(),
        planemo_option("--tutorial_title", default="Title of the tutorial", help="Title of the tutorial"),
        planemo_option("--hands_on", is_flag=True, default=True, help="Add hands-on for the new tutorial"),
        planemo_option("--slides", is_flag=True, default=False, help="Add slides for the new tutorial"),
        training_tutorial_worflow_option(),
        training_zenodo_option(),
    )


def training_init_options():
    return _compose(training_topic_option(), training_tutorial_option(), training_datatype_option())


def training_fill_data_library_options():
    return _compose(
        training_topic_name_option(),
        training_tutorial_name_req_option(),
        training_zenodo_option(),
        training_datatype_option(),
    )


def training_generate_tuto_from_wf_options():
    return _compose(
        training_topic_name_option(), training_tutorial_name_req_option(), training_tutorial_worflow_option()
    )


def shed_fail_fast_option():
    return planemo_option(
        "--fail_fast",
        is_flag=True,
        default=False,
        help="If multiple repositories are specified and an error occurs "
        "stop immediately instead of processing remaining repositories.",
    )


def lint_xsd_option():
    return planemo_option(
        "--xsd/--no_xsd", is_flag=True, default=True, help=("Include tool XSD validation in linting process.")
    )


def report_level_option():
    return planemo_option(
        "--report_level",
        type=click.Choice(["all", "warn", "error"]),
        default="all",
    )


def report_xunit():
    return planemo_option(
        "--report_xunit",
        type=click.Path(file_okay=True, resolve_path=True),
        help="Output an XUnit report, useful for CI testing",
        default=None,
    )


def skip_option():
    return planemo_option(
        "-s",
        "--skip",
        default=None,
        help=(
            "Comma-separated list of lint tests to skip (e.g. passing "
            "--skip 'citations,xml_order' would skip linting of citations "
            "and best-practice XML ordering."
        ),
    )


def fail_level_option():
    return planemo_option("--fail_level", type=click.Choice(["warn", "error"]), default="warn")


def recursive_shed_option():
    return recursive_option(
        "Recursively perform command for nested repository directories.",
    )


def recursive_option(help="Recursively perform command for subdirectories."):
    return planemo_option(
        "-r",
        "--recursive",
        is_flag=True,
        help=help,
    )


def merge_test_json():
    target_path = click.Path(
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    )
    return click.argument(
        "input_paths",
        metavar="INPUT_PATHS",
        type=target_path,
        nargs=-1,
    )


def tool_test_json(var="path"):
    target_path = click.Path(
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    )
    return click.argument(
        var,
        metavar="FILE_PATH",
        type=target_path,
        default="tool_test_output.json",
    )


def engine_options():
    return _compose(
        run_engine_option(),
        non_strict_cwl_option(),
        cwltool_no_container_option(),
        docker_galaxy_image_option(),
        docker_extra_volume_option(),
        ignore_dependency_problems_option(),
        shed_install_option(),
        install_tool_dependencies_option(),
        install_resolver_dependencies_option(),
        install_repository_dependencies_option(),
        galaxy_url_option(),
        galaxy_admin_key_option(),
        galaxy_user_key_option(),
        history_name(),
        no_wait_option(),
    )


def test_report_options():
    return _compose(
        planemo_option(
            "--test_output",
            type=click.Path(file_okay=True, resolve_path=True),
            use_global_config=True,
            default="tool_test_output.html",
            help=("Output test report (HTML - for humans) defaults to " "tool_test_output.html."),
        ),
        planemo_option(
            "--test_output_text",
            type=click.Path(file_okay=True, resolve_path=True),
            use_global_config=True,
            help=("Output test report (Basic text - for display in CI)"),
            default=None,
        ),
        planemo_option(
            "--test_output_markdown",
            type=click.Path(file_okay=True, resolve_path=True),
            use_global_config=True,
            help=("Output test report (Markdown style - for humans & " "computers)"),
            default=None,
        ),
        planemo_option(
            "--test_output_xunit",
            type=click.Path(file_okay=True, resolve_path=True),
            use_global_config=True,
            help=("Output test report (xunit style - for CI systems"),
            default=None,
        ),
        planemo_option(
            "--test_output_junit",
            type=click.Path(file_okay=True, resolve_path=True),
            use_global_config=True,
            help=("Output test report (jUnit style - for CI systems"),
            default=None,
        ),
        planemo_option(
            "--test_output_allure",
            type=click.Path(file_okay=False, resolve_path=True),
            use_global_config=True,
            help=("Output test allure2 framework resutls"),
            default=None,
        ),
    )


def profile_name_argument():
    return click.argument(
        "profile_name",
        metavar="PROFILE_NAME",
        type=str,
    )


def database_identifier_argument():
    return click.argument(
        "identifier",
        metavar="IDENTIFIER",
        type=str,
    )


def postgres_datatype_type_option():
    return planemo_option(
        "--postgres",
        "database_type",
        flag_value="postgres",
        help=("Use postgres database type."),
    )


def no_wait_option():
    return planemo_option(
        "--no_wait",
        is_flag=True,
        default=False,
        prompt=False,
        help="After invoking a job or workflow, do not wait for completion.",
    )


def database_type_option():
    return planemo_option(
        "--database_type",
        default="auto",
        type=click.Choice(
            [
                "postgres",
                "postgres_docker",
                "sqlite",
                "auto",
            ]
        ),
        use_global_config=True,
        help=(
            "Type of database to use for profile - "
            "'auto', 'sqlite', 'postgres', and 'postgres_docker' are available options. "
            "Use postgres to use an existing postgres server you user can "
            "access without a password via the psql command. Use postgres_docker "
            "to have Planemo manage a docker container running postgres. "
            "Data with postgres_docker is not yet persisted past when you restart "
            "the docker container launched by Planemo so be careful with this option."
        ),
    )


def database_source_options():
    """Database connection options for commands that utilize a database."""
    return _compose(
        planemo_option(
            "--postgres_psql_path",
            default="psql",
            use_global_config=True,
            help=("Name or or path to postgres client binary (psql)."),
        ),
        planemo_option(
            "--postgres_database_user",
            default="postgres",
            use_global_config=True,
            help=("Postgres username for managed development databases."),
        ),
        planemo_option(
            "--postgres_database_host",
            default=None,
            use_global_config=True,
            help=("Postgres host name for managed development databases."),
        ),
        planemo_option(
            "--postgres_database_port",
            default=None,
            use_global_config=True,
            help=("Postgres port for managed development databases."),
        ),
    )


def profile_database_options():
    return _compose(
        postgres_datatype_type_option(),
        database_type_option(),
        database_source_options(),
    )


def test_options():
    return _compose(
        planemo_option(
            "--update_test_data",
            is_flag=True,
            help="Update test-data directory with job outputs (normally"
            " written to directory --job_output_files if specified.)",
        ),
        paste_test_data_paths_option(),
        test_report_options(),
        planemo_option(
            "--test_output_json",
            type=click.Path(file_okay=True, resolve_path=True),
            use_global_config=True,
            help=("Output test report (planemo json) defaults to " "tool_test_output.json."),
            default="tool_test_output.json",
        ),
        planemo_option(
            "--job_output_files",
            type=click.Path(file_okay=False, resolve_path=True),
            help="Write job outputs to specified directory.",
            default=None,
        ),
        planemo_option(
            "--summary",
            type=click.Choice(["none", "minimal", "compact"]),
            default="minimal",
            help=(
                "Summary style printed to planemo's standard output (see "
                "output reports for more complete summary). Set to 'none' "
                "to disable completely."
            ),
        ),
        planemo_option(
            "--test_timeout",
            type=int,
            help="Maximum runtime of a single test in seconds.",
            default=DEFAULT_TOOL_TEST_WAIT,
        ),
    )


def _compose(*functions):
    def compose2(f, g):
        return lambda x: f(g(x))

    return functools.reduce(compose2, functions)


def dependencies_script_options():
    return _compose(
        planemo_option(
            "--download_cache",
            type=click.Path(file_okay=False, resolve_path=True),
            use_global_config=True,
            help=("Directory to cache downloaded files, default is $DOWNLOAD_CACHE"),
            default=None,
        ),
    )


def filter_exclude_option():
    return planemo_option(
        "--exclude",
        type=click.Path(resolve_path=False),
        multiple=True,
        help="Paths to exclude.",
    )


def filter_exclude_from_option():
    return planemo_option(
        "--exclude_from",
        type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
        multiple=True,
        help="File of paths to exclude.",
    )


def filter_changed_in_commit_option():
    return planemo_option(
        "--changed_in_commit_range",
        help="Exclude paths unchanged in git commit range.",
    )


def ci_chunk_count_option():
    return planemo_option(
        "--chunk_count",
        type=int,
        help="Split output into chunks of this many item and print --chunk such group.",
        default=1,
    )


def ci_group_tools_option():
    return planemo_option("--group_tools", is_flag=True, help="Group tools of the same repository on a single line.")


def ci_chunk_option():
    return planemo_option(
        "--chunk",
        type=int,
        help=("When output is split into --chunk_count groups, output the group 0-indexed" "by this option."),
        default=0,
    )


def ci_output_option():
    return planemo_option(
        "--output",
        help="File to output to, or - for standard output.",
        default="-",
    )


def ci_find_options():
    return _compose(
        filter_exclude_option(),
        filter_exclude_from_option(),
        filter_changed_in_commit_option(),
        ci_chunk_count_option(),
        ci_chunk_option(),
        ci_output_option(),
    )


def workflow_output_artifact():
    return planemo_option(
        "-o",
        "--output",
        default=None,
        type=click.Path(
            file_okay=True,
            dir_okay=False,
            readable=True,
            resolve_path=True,
        ),
    )


def tool_init_id_option(prompt=True):
    return planemo_option(
        "-i",
        "--id",
        type=click.STRING,
        prompt=prompt,
        help="Short identifier for new tool (no whitespace)",
    )


def tool_init_tool_option():
    return planemo_option(
        "-t",
        "--tool",
        default=None,
        type=click.Path(exists=False, file_okay=True, dir_okay=False, writable=True, resolve_path=True),
        help="Output path for new tool (default is <id>.xml)",
    )


def tool_init_name_option(prompt=True, help="Name for new tool (user facing)"):
    return planemo_option(
        "-n",
        "--name",
        type=click.STRING,
        prompt=prompt,
        help=help,
    )


def tool_init_version_option():
    return planemo_option(
        "--version",
        default="0.1.0",
        type=click.STRING,
        help="Tool XML version.",
    )


def tool_init_description_option():
    return planemo_option(
        "-d",
        "--description",
        type=click.STRING,
        default=None,
        prompt=False,
        help="Short description for new tool (user facing)",
    )


def tool_init_command_option():
    return planemo_option(
        "-c",
        "--command",
        type=click.STRING,
        default=None,
        prompt=False,
        help=("Command potentially including cheetah variables ()" "(e.g. 'seqtk seq -A $input > $output')"),
    )


def tool_init_doi_option():
    return planemo_option(
        "--doi",
        type=click.STRING,
        default=None,
        multiple=True,
        prompt=False,
        help=(
            "Supply a DOI (http://www.doi.org/) easing citation of the tool " "for Galxy users (e.g. 10.1101/014043)."
        ),
    )


def tool_init_test_case_option():
    return planemo_option(
        "--test_case",
        is_flag=True,
        default=None,
        prompt=False,
        help=("For use with --example_commmand, generate a tool test case from " "the supplied example."),
    )


def tool_init_macros_option():
    return planemo_option(
        "--macros",
        is_flag=True,
        default=None,
        prompt=False,
        help="Generate a macros.xml for reuse across many tools.",
    )


def tool_init_cite_url_option():
    return planemo_option(
        "--cite_url", type=click.STRING, default=None, multiple=True, prompt=False, help=("Supply a URL for citation.")
    )


def tool_init_input_option():
    return planemo_option(
        "--input",
        type=click.STRING,
        default=None,
        prompt=False,
        multiple=True,
        help="An input description (e.g. input.fasta)",
    )


def tool_init_output_option():
    return planemo_option(
        "--output",
        type=click.STRING,
        multiple=True,
        default=None,
        prompt=False,
        help=("An output location (e.g. output.bam), the Galaxy datatype is " "inferred from the extension."),
    )


def tool_init_help_text_option():
    return planemo_option(
        "--help_text",
        type=click.STRING,
        default=None,
        prompt=False,
        help="Help text (reStructuredText)",
    )


def tool_init_help_from_command_option():
    return planemo_option(
        "--help_from_command",
        type=click.STRING,
        default=None,
        prompt=False,
        help="Auto populate help from supplied command.",
    )


def tool_init_example_input_option():
    return planemo_option(
        "--example_input",
        type=click.STRING,
        default=None,
        prompt=False,
        multiple=True,
        help=("For use with --example_command, replace input file (e.g. 2.fastq " "with a data input parameter)."),
    )


def tool_init_example_output_option():
    return planemo_option(
        "--example_output",
        type=click.STRING,
        default=None,
        prompt=False,
        multiple=True,
        help=("For use with --example_command, replace input file (e.g. 2.fastq " "with a tool output)."),
    )


def tool_init_named_output_option():
    return planemo_option(
        "--named_output",
        type=click.STRING,
        multiple=True,
        default=None,
        prompt=False,
        help=(
            "Create a named output for use with command block for example "
            "specify --named_output=output1.bam and then use '-o $output1' "
            "in your command block."
        ),
    )


def tool_init_version_command_option():
    return planemo_option(
        "--version_command",
        type=click.STRING,
        default=None,
        prompt=False,
        help="Command to print version (e.g. 'seqtk --version')",
    )


REQUIREMENT_HELP = "Add a tool requirement package (e.g. 'seqtk' or 'seqtk@1.68')."


def tool_init_requirement_option(help=REQUIREMENT_HELP):
    return planemo_option(
        "--requirement",
        type=click.STRING,
        default=None,
        multiple=True,
        prompt=False,
        help=help,
    )


def tool_init_container_option():
    return planemo_option(
        "--container",
        type=click.STRING,
        default=None,
        multiple=True,
        prompt=False,
        help="Add a Docker image identifier for this tool.",
    )


EXAMPLE_COMMAND_HELP = (
    "Example to command with paths to build Cheetah template from "
    "(e.g. 'seqtk seq -A 2.fastq > 2.fasta'). Option cannot be used "
    "with --command, should be used --example_input and "
    "--example_output."
)


def tool_init_example_command_option(help=EXAMPLE_COMMAND_HELP):
    return planemo_option(
        "--example_command",
        type=click.STRING,
        default=None,
        prompt=False,
        help=help,
    )


def mulled_conda_option():
    return planemo_option(
        "--mulled_conda_version",
        type=click.STRING,
        default=None,
        help=(
            "Install a specific version of Conda before running the command, by "
            "default the version that comes with the continuumio miniconda3 image "
            "will be used under Linux and under Mac OS X Conda will be upgraded to "
            "to work around a bug in 4.2."
        ),
    )


def mulled_namespace_option():
    return planemo_option(
        "--mulled_namespace",
        type=click.STRING,
        default="biocontainers",
        help=(
            "Build a mulled image with the specified namespace - defaults to "
            "biocontainers. Galaxy currently only recognizes images with the "
            "namespace biocontainers."
        ),
    )


def mulled_action_option():
    return planemo_option(
        "--mulled_command",
        type=click.STRING,
        default="build-and-test",
        help=("Mulled action to perform for targets - this defaults to 'build-and-test'."),
    )


def mulled_options():
    return _compose(
        mulled_conda_option(),
        mulled_namespace_option(),
        mulled_action_option(),
    )



"""Describe artifacts that can be run, tested, and linted."""


import abc
import os
from distutils.dir_util import copy_tree
from enum import (
    auto,
    Enum,
)
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    overload,
    Union,
)
from urllib.parse import urlparse

import yaml
from bioblend.galaxy import GalaxyInstance
from galaxy.tool_util.cwl.parser import workflow_proxy
from galaxy.tool_util.loader_directory import (
    is_a_yaml_with_class,
    looks_like_a_cwl_artifact,
    looks_like_a_data_manager_xml,
    looks_like_a_tool_cwl,
    looks_like_a_tool_xml,
)
from galaxy.tool_util.parser import get_tool_source
from typing_extensions import Literal

from planemo.exit_codes import (
    EXIT_CODE_UNKNOWN_FILE_TYPE,
    ExitCodeException,
)
from planemo.galaxy.workflows import (
    describe_outputs,
    GALAXY_WORKFLOWS_PREFIX,
    WorkflowOutput,
)
from planemo.io import error
from planemo.shed import DOCKSTORE_REGISTRY_CONF
from planemo.test import (
    check_output,
    for_collections,
)

TEST_SUFFIXES = ["-tests", "_tests", "-test", "_test"]
TEST_EXTENSIONS = [".yml", ".yaml", ".json"]

TEST_FILE_NOT_LIST_MESSAGE = "Invalid test definition file [%s] - file must " "contain a list of tests"
TEST_FIELD_MISSING_MESSAGE = "Invalid test definition [test #%d in %s] -" "defintion must field [%s]."
GALAXY_TOOLS_PREFIX = "gxid://tools/"


class RunnableType(Enum):
    galaxy_tool = auto()
    galaxy_datamanager = auto()
    galaxy_workflow = auto()
    cwl_tool = auto()
    cwl_workflow = auto()
    directory = auto()

    @property
    def has_tools(runnable_type):
        return runnable_type.name in ["galaxy_tool", "galaxy_datamanager", "cwl_tool", "directory"]

    @property
    def is_single_artifact(runnable_type):
        return runnable_type.name not in ["directory"]

    @property
    def test_data_in_parent_dir(runnable_type):
        return runnable_type.name in ["galaxy_datamanager"]

    @property
    def is_galaxy_artifact(runnable_type) -> bool:
        return "galaxy" in runnable_type.name

    @property
    def is_cwl_artifact(runnable_type) -> bool:
        return "cwl" in runnable_type.name


class Runnable(NamedTuple):
    """Abstraction describing tools and workflows."""

    uri: str
    type: RunnableType

    @property
    def path(self) -> str:
        uri = self.uri
        if self.is_remote_workflow_uri:
            parse_result = urlparse(uri)
            query = parse_result.query
            if query:
                assert query.startswith("runnable_path=")
                return query[len("runnable_path=") :]
            else:
                raise ValueError(f"Runnable with URI {uri} is remote resource without local path")
        else:
            return uri

    @property
    def has_path(self):
        try:
            self.path
            return True
        except ValueError:
            return False

    @property
    def is_remote_workflow_uri(self) -> bool:
        return self.uri.startswith(GALAXY_WORKFLOWS_PREFIX)

    @property
    def test_data_search_path(self) -> str:
        """During testing, path to search for test data files."""
        if self.type.name in ["galaxy_datamanager"]:
            return os.path.join(os.path.dirname(self.path), os.path.pardir)
        else:
            return self.path

    @property
    def tool_data_search_path(self) -> str:
        """During testing, path to search for Galaxy tool data tables."""
        return self.test_data_search_path

    @property
    def data_manager_conf_path(self) -> Optional[str]:
        """Path of a Galaxy data manager configuration for runnable or None."""
        if self.type.name in ["galaxy_datamanager"]:
            return os.path.join(os.path.dirname(self.path), os.pardir, "data_manager_conf.xml")
        return None

    @property
    def has_tools(self) -> property:
        """Boolean indicating if this runnable corresponds to one or more tools."""
        return _runnable_delegate_attribute("has_tools")

    @property
    def is_single_artifact(self) -> property:
        """Boolean indicating if this runnable is a single artifact.

        Currently only directories are considered not a single artifact.
        """
        return _runnable_delegate_attribute("is_single_artifact")


class Rerunnable(NamedTuple):
    """Abstraction describing artifacts (histories, invocation, jobs) on external Galaxy instances with associated rerunnable and remappable jobs."""

    rerunnable_id: str
    rerunnable_type: str
    server_url: str


def _runnable_delegate_attribute(attribute: str) -> property:
    def getter(runnable):
        return getattr(runnable.type, attribute)

    return property(getter)


def _copy_runnable_tree(path: str, runnable_type: RunnableType, temp_path: str) -> str:
    dir_to_copy = None
    if runnable_type in {RunnableType.galaxy_tool, RunnableType.cwl_tool}:
        dir_to_copy = os.path.dirname(path)
        path = os.path.join(temp_path, os.path.basename(path))
    elif runnable_type == RunnableType.directory:
        dir_to_copy = path
        path = temp_path
    elif runnable_type == RunnableType.galaxy_datamanager:
        dir_to_copy = os.path.join(os.path.dirname(path), os.pardir)
        path_to_data_manager_tool = os.path.relpath(path, dir_to_copy)
        path = os.path.join(temp_path, path_to_data_manager_tool)
    if dir_to_copy:
        copy_tree(dir_to_copy, temp_path, update=True)
    return path


def workflows_from_dockstore_yaml(path):
    workflows = []
    parent_dir = Path(path).absolute().parent
    with open(path) as y:
        for workflow in yaml.safe_load(y).get("workflows", []):
            workflow_path = workflow.get("primaryDescriptorPath")
            if workflow_path:
                if workflow_path.startswith("/"):
                    workflow_path = workflow_path[1:]
            workflows.append(parent_dir.joinpath(workflow_path))
    return workflows


def workflow_dir_runnables(path: str, return_all: bool = False) -> Optional[Union[Runnable, List[Runnable]]]:
    dockstore_path = os.path.join(path, DOCKSTORE_REGISTRY_CONF)
    if os.path.exists(dockstore_path):
        runnables = [
            Runnable(str(path), RunnableType.galaxy_workflow) for path in workflows_from_dockstore_yaml(dockstore_path)
        ]
        if return_all:
            return runnables
        else:
            return runnables[0]
    return None


@overload
def for_path(path: str, temp_path: Optional[str] = None, return_all: Literal[False] = False) -> Runnable:
    ...


@overload
def for_path(path: str, temp_path: Optional[str] = None, return_all: bool = False) -> Union[Runnable, List[Runnable]]:
    pass


def for_path(path: str, temp_path: Optional[str] = None, return_all: bool = False) -> Union[Runnable, List[Runnable]]:
    """Produce a class:`Runnable` for supplied path."""
    runnable_type = None
    if os.path.isdir(path):
        runnable = workflow_dir_runnables(path, return_all=return_all)
        if runnable:
            return runnable
        runnable_type = RunnableType.directory
    elif looks_like_a_tool_cwl(path):
        runnable_type = RunnableType.cwl_tool
    elif looks_like_a_data_manager_xml(path):
        runnable_type = RunnableType.galaxy_datamanager
    elif looks_like_a_tool_xml(path):
        runnable_type = RunnableType.galaxy_tool
    elif is_a_yaml_with_class(path, ["GalaxyWorkflow"]):
        runnable_type = RunnableType.galaxy_workflow
    elif path.endswith(".ga"):
        runnable_type = RunnableType.galaxy_workflow
    elif looks_like_a_cwl_artifact(path, ["Workflow"]):
        runnable_type = RunnableType.cwl_workflow
    else:
        # Check to see if it is a Galaxy workflow with a different extension
        try:
            with open(path) as f:
                as_dict = yaml.safe_load(f)
            if as_dict.get("a_galaxy_workflow", False):
                runnable_type = RunnableType.galaxy_workflow
        except Exception:
            pass

    if runnable_type is None:
        error(f"Unable to determine runnable type for path [{path}]")
        raise ExitCodeException(EXIT_CODE_UNKNOWN_FILE_TYPE)

    if temp_path:
        path = _copy_runnable_tree(path, runnable_type, temp_path)

    return Runnable(path, runnable_type)


def for_paths(paths: Iterable[str], temp_path: Optional[str] = None) -> List[Runnable]:
    """Return a specialized list of Runnable objects for paths."""
    return [for_path(path, temp_path=temp_path) for path in paths]


def for_uri(uri: str) -> Runnable:
    """Produce a class:`Runnable` for supplied Galaxy workflow or tool ID."""
    runnable_type = RunnableType.galaxy_tool if uri.startswith(GALAXY_TOOLS_PREFIX) else RunnableType.galaxy_workflow
    runnable = Runnable(uri, runnable_type)
    return runnable


def cases(runnable: Runnable) -> List["AbstractTestCase"]:
    """Build a `list` of :class:`TestCase` objects for specified runnable."""
    cases: List["AbstractTestCase"] = []

    tests_path = _tests_path(runnable)
    if tests_path is None:
        if runnable.type in (RunnableType.galaxy_tool, RunnableType.galaxy_datamanager):
            if runnable.uri.startswith(GALAXY_TOOLS_PREFIX):
                return [DelayedGalaxyToolTestCase(runnable)]
            tool_source = get_tool_source(runnable.path)
            test_dicts = tool_source.parse_tests_to_dict()
            tool_id = tool_source.parse_id()
            tool_version = tool_source.parse_version()
            for i, test_dict in enumerate(test_dicts.get("tests", [])):
                cases.append(ExternalGalaxyToolTestCase(runnable, tool_id, tool_version, i, test_dict))
        return cases

    tests_directory = os.path.abspath(os.path.dirname(tests_path))

    def normalize_to_tests_path(path: str) -> str:
        if not os.path.isabs(path):
            absolute_path = os.path.join(tests_directory, path)
        else:
            absolute_path = path
        return os.path.normpath(absolute_path)

    with open(tests_path) as f:
        tests_def = yaml.safe_load(f)

    if not isinstance(tests_def, list):
        message = TEST_FILE_NOT_LIST_MESSAGE % tests_path
        raise Exception(message)

    for i, test_def in enumerate(tests_def):
        if "job" not in test_def:
            message = TEST_FIELD_MISSING_MESSAGE % (i + 1, tests_path, "job")
            raise Exception(message)
        job_def = test_def["job"]
        if isinstance(job_def, dict):
            job_path = None
            job = job_def
        else:
            job_path = normalize_to_tests_path(job_def)
            job = None

        doc = test_def.get("doc")
        output_expectations = test_def.get("outputs", {})
        case = TestCase(
            runnable=runnable,
            tests_directory=tests_directory,
            output_expectations=output_expectations,
            index=i,
            job_path=job_path,
            job=job,
            doc=doc,
        )
        cases.append(case)

    return cases


class AbstractTestCase(metaclass=abc.ABCMeta):
    """Description of a test case for a runnable."""

    def structured_test_data(self, run_response):
        """Result of executing this test case - a "structured_data" dict.

        :rtype: dict
        :return:
                 For example::

                   {
                       "id": "",
                       "has_data": true,
                       "data": {
                           "status": "success", // error, skip,
                           "job": {
                               "command_line": "cat moo",
                               "stdout": "",
                               "stderr": ""
                           },
                           "output_problems": [],
                           "execution_problem": "",
                           "inputs" = {},
                           "problem_log": ""
                       }
                   }
        """


class TestCase(AbstractTestCase):
    """Describe an abstract test case for a specified runnable."""

    def __init__(
        self,
        runnable: Runnable,
        tests_directory: str,
        output_expectations: Dict[str, Any],
        job_path: Optional[str],
        job: Optional[Dict],
        index: int,
        doc: Optional[str],
    ) -> None:
        """Construct TestCase object from required attributes."""
        self.runnable = runnable
        self.job_path = job_path
        self.job = job
        self.output_expectations = output_expectations
        self.tests_directory = tests_directory
        self.index = index
        self.doc = doc

    def __repr__(self) -> str:
        return f"TestCase ({self.doc}) for runnable ({self.runnable}) with job ({self.job}) and expected outputs ({self.output_expectations}) in directory ({self.tests_directory}) with id ({self.index})"

    def structured_test_data(self, run_response: "RunResponse") -> Dict[str, Any]:
        """Check a test case against outputs dictionary."""
        return run_response.structured_data(self)

    @property
    def _job(self):
        if self.job_path is not None:
            with open(self.job_path) as f:
                return f.read()
        else:
            return self.job

    @property
    def input_ids(self) -> List[str]:
        """Labels of inputs specified in test description."""
        return list(self._job.keys())

    @property
    def tested_output_ids(self) -> List[str]:
        """Labels of outputs checked in test description."""
        return list(self.output_expectations.keys())

    def _check_output(
        self,
        output_id: str,
        output_value: Any,
        output_test: Any,
    ) -> List[str]:
        output_problems = []
        if not isinstance(output_test, dict):
            if output_test != output_value:
                message = f"Output [{output_id}] value [{output_value}] does not match expected value [{output_test}]."
                output_problems.append(message)
        else:
            if not for_collections(output_test):
                if not isinstance(output_value, dict):
                    message = f"Expected file properties for output [{output_id}]"
                    print(message)
                    print(output_value)
                    output_problems.append(message)
                    return output_problems
                if "path" not in output_value and "location" in output_value:
                    assert output_value["location"].startswith("file://")
                    output_value["path"] = output_value["location"][len("file://") :]
                if "path" not in output_value:
                    message = f"No path specified for expected output file [{output_id}]"
                    output_problems.append(message)
                    print(message)
                    return output_problems
            else:
                output_test["name"] = output_id

            output_problems.extend(
                check_output(
                    self.runnable,
                    output_value,
                    output_test,
                    # TODO: needs kwds in here...
                )
            )

        return output_problems

    @property
    def _test_id(self) -> str:
        if self.runnable.type in [
            RunnableType.cwl_tool,
            RunnableType.galaxy_tool,
        ]:
            return get_tool_source(self.runnable.path).parse_id()
        else:
            return os.path.basename(self.runnable.path)


class ExternalGalaxyToolTestCase(AbstractTestCase):
    """Special class of AbstractCase that doesn't use job_path but uses test data from a Galaxy server."""

    def __init__(
        self,
        runnable: Runnable,
        tool_id: Optional[str],
        tool_version: Optional[str],
        test_index: Optional[int],
        test_dict: Any,
    ) -> None:
        """Construct TestCase object from required attributes."""
        self.runnable = runnable
        self.tool_id = tool_id
        self.tool_version = tool_version
        self.test_index = test_index
        self.test_dict = test_dict

    def structured_test_data(self, run_response: Dict[str, Any]) -> Dict[str, Any]:
        """Just return the structured_test_data generated from galaxy-tool-util for this test variant."""
        return run_response


class DelayedGalaxyToolTestCase(ExternalGalaxyToolTestCase):
    """Special class that requires installing tools prior to finding test cases."""

    def __init__(self, runnable: Runnable) -> None:
        super().__init__(runnable, tool_id=None, tool_version=None, test_index=None, test_dict=None)


def _tests_path(runnable: Runnable) -> Optional[str]:
    if not runnable.is_single_artifact:
        raise NotImplementedError("Tests for directories are not yet implemented.")

    runnable_path = runnable.path
    base, _ = os.path.splitext(runnable_path)

    for test_suffix in TEST_SUFFIXES:
        for test_extension in TEST_EXTENSIONS:
            test_path = base + test_suffix + test_extension
            if os.path.exists(test_path):
                return test_path

    return None


def get_outputs(runnable: Runnable, gi: Optional[GalaxyInstance] = None) -> List["RunnableOutput"]:
    """Return a list of :class:`RunnableOutput` objects for this runnable.

    Supply bioblend user Galaxy instance object (as gi) if additional context
    needed to resolve workflow details.
    """
    if not runnable.is_single_artifact:
        raise NotImplementedError("Cannot generate outputs for a directory.")
    if runnable.type in [RunnableType.galaxy_tool, RunnableType.cwl_tool]:
        tool_source = get_tool_source(runnable.path)
        # TODO: do something with collections at some point
        output_datasets, _ = tool_source.parse_outputs(None)
        return [ToolOutput(o) for o in output_datasets.values()]
    elif runnable.type == RunnableType.galaxy_workflow:
        workflow_outputs = describe_outputs(runnable, gi=gi)
        return [GalaxyWorkflowOutput(o) for o in workflow_outputs]
    elif runnable.type == RunnableType.cwl_workflow:
        workflow = workflow_proxy(runnable.path, strict_cwl_validation=False)
        return [CwlWorkflowOutput(label) for label in workflow.output_labels]
    else:
        raise NotImplementedError("Getting outputs for this artifact type is not yet supported.")


class RunnableOutput(metaclass=abc.ABCMeta):
    """Description of a single output of an execution of a Runnable."""

    @abc.abstractproperty
    def get_id(self):
        """An identifier that describes this output."""

    def is_optional(self):
        return False


class ToolOutput(RunnableOutput):
    """Implementation of RunnableOutput corresponding to Galaxy tool outputs."""

    def __init__(self, tool_output):
        self._tool_output = tool_output

    def get_id(self):
        return self._tool_output.name


class GalaxyWorkflowOutput(RunnableOutput):
    """Implementation of RunnableOutput corresponding to Galaxy workflow outputs."""

    def __init__(self, workflow_output: WorkflowOutput) -> None:
        self._workflow_output = workflow_output

    def get_id(self) -> Optional[str]:
        return self._workflow_output.label

    def is_optional(self):
        return self.workflow_output.optional

    @property
    def workflow_output(self):
        return self._workflow_output


class CwlWorkflowOutput(RunnableOutput):
    """Implementation of RunnableOutput corresponding to CWL outputs."""

    def __init__(self, label: str) -> None:
        self._label = label

    def get_id(self) -> str:
        return self._label


class RunResponse(metaclass=abc.ABCMeta):
    """Description of an attempt for an engine to execute a Runnable."""

    @property
    def start_datetime(self) -> None:
        """Start datetime of run."""
        return None

    @property
    def end_datetime(self) -> None:
        """End datetime of run."""
        return None

    @abc.abstractproperty
    def was_successful(self):
        """Indicate whether an error was encountered while executing this runnable.

        If successful, response should conform to the SuccessfulRunResponse interface,
        otherwise it will conform to the ErrorRunResponse interface.
        """

    @abc.abstractproperty
    def job_info(self):
        """If job information is available, return as dictionary."""

    @abc.abstractproperty
    def invocation_details(self):
        """If workflow invocation details are available, return as dictionary."""

    @abc.abstractproperty
    def log(self):
        """If engine related log is available, return as text data."""

    def structured_data(self, test_case: Optional[TestCase] = None) -> Dict[str, Any]:
        output_problems = []
        if isinstance(self, SuccessfulRunResponse) and self.was_successful:
            outputs_dict = self.outputs_dict
            execution_problem = None
            if test_case:
                for output_id, output_test in test_case.output_expectations.items():
                    if output_id not in outputs_dict:
                        message = f"Expected output [{output_id}] not found in results."
                        output_problems.append(message)
                        continue

                    output_value = outputs_dict[output_id]
                    output_problems.extend(test_case._check_output(output_id, output_value, output_test))
            if output_problems:
                status = "failure"
            else:
                status = "success"
        else:
            execution_problem = getattr(self, "error_message", None)
            status = "error"
        data_dict: Dict[str, Any] = dict(status=status)
        if status != "success":
            data_dict["output_problems"] = output_problems
            data_dict["execution_problem"] = execution_problem
        log = self.log
        if log is not None:
            data_dict["problem_log"] = log
        job_info = self.job_info
        if job_info is not None:
            data_dict["job"] = job_info
        invocation_details = self.invocation_details
        if invocation_details is not None:
            data_dict["invocation_details"] = invocation_details
        if self.start_datetime is not None:
            data_dict["start_datetime"] = self.start_datetime.isoformat()
        if self.end_datetime is not None:
            data_dict["end_datetime"] = self.end_datetime.isoformat()
        if test_case:
            data_dict["inputs"] = test_case._job
            return dict(
                id=(f"{test_case._test_id}_{test_case.index}"),
                has_data=True,
                data=data_dict,
                doc=test_case.doc,
                test_type=test_case.runnable.type.name,
            )
        else:
            assert isinstance(self, SuccessfulRunResponse)
            return dict(
                id=self._runnable.uri,
                has_data=True,
                data=data_dict,
                doc=None,
                test_type=self._runnable.type.name,
            )


class SuccessfulRunResponse(RunResponse, metaclass=abc.ABCMeta):
    """Description of the results of an engine executing a Runnable."""

    def __init__(self, runnable: "Runnable") -> None:
        self._runnable = runnable

    def was_successful(self):
        """Return `True` to indicate this run was successful."""
        return True

    @abc.abstractproperty
    def outputs_dict(self):
        """Return a dict of output descriptions."""


class ErrorRunResponse(RunResponse):
    """Description of an error while attempting to execute a Runnable."""

    def __init__(
        self, error_message, job_info=None, invocation_details=None, log=None, start_datetime=None, end_datetime=None
    ):
        """Create an ErrorRunResponse with specified error message."""
        self._error_message = error_message
        self._job_info = job_info
        self._invocation_details = invocation_details
        self._log = log
        self._start_datetime = start_datetime
        self._end_datetime = end_datetime

    @property
    def start_datetime(self):
        """Start datetime of run."""
        return self._start_datetime

    @property
    def end_datetime(self):
        """End datetime of run."""
        return self._end_datetime

    @property
    def error_message(self):
        """Error message describing the problem with execution of the runnable."""
        return self._error_message

    @property
    def was_successful(self):
        """Return `False` to indicate this run was successful."""
        return False

    @property
    def job_info(self):
        """Return potentially null stored `job_info` dict."""
        return self._job_info

    @property
    def invocation_details(self):
        return self._invocation_details

    @property
    def log(self):
        """Return potentially null stored `log` text."""
        return self._log

    def __str__(self):
        """Print a helpful error description of run."""
        message = f"Run failed with message [{self.error_message}]"
        log = self.log
        if log:
            message += f" and log [{log}]"
        return message


__all__ = (
    "cases",
    "ErrorRunResponse",
    "for_path",
    "for_paths",
    "get_outputs",
    "Runnable",
    "RunnableType",
    "RunResponse",
    "RunnableOutput",
    "SuccessfulRunResponse",
    "TestCase",
)



"""Module describing the planemo ``training_fill_data_library`` command."""
import click

from planemo import options
from planemo.cli import command_function
from planemo.training import Training


@click.command("training_fill_data_library")
@options.optional_tools_arg(multiple=True, allow_uris=True)
@options.training_fill_data_library_options()
@command_function
def cli(ctx, uris, **kwds):
    """Build training template from workflow."""
    kwds["no_dependency_resolution"] = True
    training = Training(kwds)
    training.fill_data_library(ctx)



"""Module describing the planemo ``conda_init`` command."""
import click
from galaxy.tool_util.deps import conda_util

from planemo import options
from planemo.cli import command_function
from planemo.conda import build_conda_context
from planemo.exit_codes import EXIT_CODE_ALREADY_EXISTS
from planemo.io import (
    info,
    warn,
)

MESSAGE_ERROR_ALREADY_EXISTS = "conda_init failed - Conda appears to already be installed at '%s'"
MESSAGE_ERROR_FAILED = "conda_init failed - failed to install to '%s'"
MESSAGE_INSTALL_OKAY = "Conda installation succeeded - Conda is available at '%s'"


@click.command("conda_init")
@options.conda_target_options(include_local=False)  # Always use local during init.
@command_function
def cli(ctx, **kwds):
    """Download and install conda.

    This will download conda for managing dependencies for your platform
    using the appropriate Miniconda installer.

    By running this command, you are agreeing to the terms of the conda
    license a 3-clause BSD 3 license. Please review full license at
    http://docs.continuum.io/anaconda/eula.

    Planemo will print a warning and terminate with an exit code of 7
    if Conda is already installed.
    """
    conda_context = build_conda_context(ctx, **kwds)
    if conda_context.is_conda_installed():
        warn(MESSAGE_ERROR_ALREADY_EXISTS % conda_context.conda_exec)
        exit = EXIT_CODE_ALREADY_EXISTS
    else:
        exit = conda_util.install_conda(conda_context=conda_context, force_conda_build=True)
        if exit:
            warn(MESSAGE_ERROR_FAILED % conda_context.conda_exec)
        else:
            info(MESSAGE_INSTALL_OKAY % conda_context.conda_exec)

    ctx.exit(exit)



"""Module describing the planemo ``open`` command."""
import click

from planemo.cli import command_function


@click.command("docs")
@click.argument(
    "path",
    metavar="PATH",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    ),
    default="tool_test_output.html",
)
@command_function
def cli(ctx, path, **kwds):
    """Open latest Planemo test results in a web browser."""
    click.launch(path)



"""Describe the interface classes of the planemo.database package."""

import abc


class DatabaseSource(metaclass=abc.ABCMeta):
    """Interface describing a source of profile databases."""

    @abc.abstractmethod
    def create_database(self, identifier):
        """Create a database with specified short identifier.

        Throw an exception if it already exists.
        """

    @abc.abstractmethod
    def delete_database(self, identifier):
        """Delete a database with specified short identifier.

        Throw an exception if it already exists.
        """

    @abc.abstractmethod
    def list_databases(self):
        """Return identifiers associated with database source."""

    @abc.abstractmethod
    def sqlalchemy_url(self, identifier):
        """Return a URL string for use by sqlalchemy."""


__all__ = ("DatabaseSource",)



"""Module describing the planemo ``ci_find_tools`` command."""

import click

from planemo import options
from planemo.ci import (
    filter_paths,
    group_paths,
    print_path_list,
)
from planemo.cli import command_function
from planemo.tools import (
    is_tool_load_error,
    yield_tool_sources_on_paths,
)


@click.command("ci_find_tools")
@options.shed_project_arg()
@options.ci_find_options()
@options.ci_group_tools_option()
@command_function
def cli(ctx, paths, **kwds):
    """Find all tools in one or more directories.

    Tools can be chunked up, filtered, etc... to build lists of tools to perform
    operations over for continuous integration operations.
    """
    tool_paths = []
    for (tool_path, tool_source) in yield_tool_sources_on_paths(ctx, paths, recursive=True):
        if is_tool_load_error(tool_source):
            continue
        tool_paths.append(tool_path)

    paths = filter_paths(ctx, tool_paths, path_type="file", **kwds)
    if kwds.get("group_tools", False):
        paths = group_paths(paths)
    print_path_list(paths, **kwds)



"""This modules describes the abstraction of a Galaxy profile.

This is a workspace with a specific default configuration and shed
tool setup. It is meant to be used with various serve commands.
"""
import json
import os
import shutil

from galaxy.util.commands import which

from planemo.database import create_database_source
from planemo.galaxy.api import test_credentials_valid
from .config import DATABASE_LOCATION_TEMPLATE

PROFILE_OPTIONS_JSON_NAME = "planemo_profile_options.json"
ALREADY_EXISTS_EXCEPTION = "Cannot create profile with name [%s], directory [%s] already exists."


def profile_exists(ctx, profile_name, **kwds):
    """Return a truthy value iff the specified profile already exists."""
    profile_directory = _profile_directory(ctx, profile_name)
    return os.path.exists(profile_directory)


def list_profiles(ctx, **kwds):
    """Return a list of current profile names."""
    return os.listdir(ctx.galaxy_profiles_directory)


def delete_profile(ctx, profile_name, **kwds):
    """Delete profile with the specified name."""
    profile_directory = _profile_directory(ctx, profile_name)
    profile_options = _read_profile_options(profile_directory)
    profile_options, profile_options_path = _load_profile_to_json(ctx, profile_name)
    if profile_options["engine"] != "external_galaxy":
        database_type = profile_options.get("database_type")
        kwds["database_type"] = database_type
        if database_type != "sqlite":
            database_source = create_database_source(**kwds)
            database_identifier = _profile_to_database_identifier(profile_name)
            database_source.delete_database(
                database_identifier,
            )
    shutil.rmtree(profile_directory)


def create_profile(ctx, profile_name, **kwds):
    """Create a profile with the specified name."""
    engine_type = kwds.get("engine", "galaxy")
    profile_directory = _profile_directory(ctx, profile_name)
    if profile_exists(ctx, profile_name, **kwds):
        message = ALREADY_EXISTS_EXCEPTION % (profile_name, profile_directory)
        raise Exception(message)

    os.makedirs(profile_directory)

    if engine_type == "docker_galaxy":
        create_for_engine = _create_profile_docker
    elif engine_type == "external_galaxy" or kwds.get("galaxy_url"):
        create_for_engine = _create_profile_external
    else:
        create_for_engine = _create_profile_local

    stored_profile_options = create_for_engine(ctx, profile_directory, profile_name, kwds)

    profile_options_path = _stored_profile_options_path(profile_directory)
    with open(profile_options_path, "w") as f:
        json.dump(stored_profile_options, f)


def _create_profile_docker(ctx, profile_directory, profile_name, kwds):
    export_directory = os.path.join(profile_directory, "export")
    os.makedirs(export_directory)
    return {
        "engine": "docker_galaxy",
    }


def _create_profile_local(ctx, profile_directory, profile_name, kwds):
    database_type = kwds.get("database_type", "auto")
    if database_type == "auto":
        if which("psql"):
            database_type = "postgres"
        elif which("docker"):
            database_type = "postgres_docker"
        else:
            database_type = "sqlite"

    if database_type != "sqlite":
        database_source = create_database_source(**kwds)
        database_identifier = _profile_to_database_identifier(profile_name)
        database_source.create_database(
            database_identifier,
        )
        database_connection = database_source.sqlalchemy_url(database_identifier)
    else:
        database_location = os.path.join(profile_directory, "galaxy.sqlite")
        database_connection = DATABASE_LOCATION_TEMPLATE % database_location

    return {
        "database_type": database_type,
        "database_connection": database_connection,
        "engine": "galaxy",
    }


def _create_profile_external(ctx, profile_directory, profile_name, kwds):
    url = kwds.get("galaxy_url")
    api_key = kwds.get("galaxy_admin_key") or kwds.get("galaxy_user_key")
    if test_credentials_valid(url=url, key=api_key, is_admin=kwds.get("galaxy_admin_key")):
        return {
            "galaxy_url": url,
            "galaxy_user_key": kwds.get("galaxy_user_key"),
            "galaxy_admin_key": kwds.get("galaxy_admin_key"),
            "engine": "external_galaxy",
        }
    else:
        raise ConnectionError("The credentials provided for an external Galaxy instance are not valid.")


def ensure_profile(ctx, profile_name, **kwds):
    """Ensure a Galaxy profile exists and return profile defaults."""
    if not profile_exists(ctx, profile_name, **kwds):
        create_profile(ctx, profile_name, **kwds)

    return _profile_options(ctx, profile_name, **kwds)


def create_alias(ctx, alias, obj, profile_name, **kwds):
    profile_options, profile_options_path = _load_profile_to_json(ctx, profile_name)

    if profile_options.get("aliases"):
        profile_options["aliases"][alias] = obj
    else:  # no aliases yet defined
        profile_options["aliases"] = {alias: obj}

    with open(profile_options_path, "w") as f:
        json.dump(profile_options, f)

    return 0


def list_alias(ctx, profile_name, **kwds):
    profile_options, _ = _load_profile_to_json(ctx, profile_name)
    return profile_options.get("aliases", {})


def delete_alias(ctx, alias, profile_name, **kwds):
    profile_options, profile_options_path = _load_profile_to_json(ctx, profile_name)
    if alias not in profile_options.get("aliases", {}):
        return 1
    else:
        del profile_options["aliases"][alias]

    with open(profile_options_path, "w") as f:
        json.dump(profile_options, f)

    return 0


def translate_alias(ctx, alias, profile_name):
    if not profile_name:
        return alias
    aliases = _load_profile_to_json(ctx, profile_name)[0].get("aliases", {})
    return aliases.get(alias, alias)


def _load_profile_to_json(ctx, profile_name):
    if not profile_exists(ctx, profile_name):
        raise Exception("That profile does not exist. Create it with `planemo profile_create`")
    profile_directory = _profile_directory(ctx, profile_name)
    profile_options_path = _stored_profile_options_path(profile_directory)
    with open(profile_options_path) as f:
        profile_options = json.load(f)
    return profile_options, profile_options_path


def _profile_options(ctx, profile_name, **kwds):
    profile_directory = _profile_directory(ctx, profile_name)
    profile_options = _read_profile_options(profile_directory)

    if profile_options["engine"] == "docker_galaxy":
        engine_options = dict(export_directory=os.path.join(profile_directory, "export"))
    else:
        file_path = os.path.join(profile_directory, "files")
        shed_tool_path = os.path.join(profile_directory, "shed_tools")
        shed_tool_conf = os.path.join(profile_directory, "shed_tool_conf.xml")
        tool_dependency_dir = os.path.join(profile_directory, "deps")

        engine_options = dict(
            file_path=file_path,
            tool_dependency_dir=tool_dependency_dir,
            shed_tool_conf=shed_tool_conf,
            shed_tool_path=shed_tool_path,
            galaxy_brand=profile_name,
        )
    profile_options.update(engine_options)
    profile_options["galaxy_brand"] = profile_name
    return profile_options


def _profile_to_database_identifier(profile_name):
    char_lst = [c if c.isalnum() else "_" for c in profile_name]
    return "plnmoprof_%s" % "".join(char_lst)


def _read_profile_options(profile_directory):
    profile_options_path = _stored_profile_options_path(profile_directory)
    with open(profile_options_path) as f:
        profile_options = json.load(f)
    return profile_options


def _stored_profile_options_path(profile_directory):
    profile_options_path = os.path.join(profile_directory, PROFILE_OPTIONS_JSON_NAME)
    return profile_options_path


def _profile_directory(ctx, profile_name):
    return os.path.join(ctx.galaxy_profiles_directory, profile_name)


__all__ = (
    "create_profile",
    "delete_profile",
    "ensure_profile",
    "list_profiles",
)



"""Module describing the planemo ``list_invocations`` command."""
import json

import click

from planemo import options
from planemo.cli import command_function
from planemo.galaxy import profiles
from planemo.galaxy.api import get_invocations
from planemo.galaxy.workflows import remote_runnable_to_workflow_id
from planemo.io import (
    error,
    info,
)
from planemo.runnable_resolve import for_runnable_identifier

try:
    from tabulate import tabulate
except ImportError:
    tabulate = None  # type: ignore


@click.command("list_invocations")
@click.argument(
    "workflow_identifier",
    type=click.STRING,
)
@options.profile_option(required=True)
@command_function
def cli(ctx, workflow_identifier, **kwds):
    """
    Get a list of invocations for a particular workflow ID or alias.
    """
    info(f"Looking for invocations for workflow {workflow_identifier}...")
    runnable = for_runnable_identifier(ctx, workflow_identifier, kwds)
    profile = profiles.ensure_profile(ctx, kwds.get("profile"))
    assert runnable.is_remote_workflow_uri
    workflow_id = remote_runnable_to_workflow_id(runnable)

    invocations = get_invocations(
        url=profile["galaxy_url"],
        key=profile["galaxy_admin_key"] or profile["galaxy_user_key"],
        workflow_id=workflow_id,
    )
    if tabulate:
        state_colors = {
            "ok": "\033[92m",  # green
            "running": "\033[93m",  # yellow
            "error": "\033[91m",  # red
            "paused": "\033[96m",  # cyan
            "deleted": "\033[95m",  # magenta
            "deleted_new": "\033[95m",  # magenta
            "new": "\033[96m",  # cyan
            "queued": "\033[93m",  # yellow
        }
        print(
            tabulate(
                {
                    "Invocation ID": invocations.keys(),
                    "Jobs status": [
                        ", ".join([f"{state_colors[k]}{v} jobs {k}\033[0m" for k, v in inv["states"].items()])
                        for inv in invocations.values()
                    ],
                    "Invocation report URL": [
                        "{}/workflows/invocations/report?id={}".format(profile["galaxy_url"].strip("/"), inv_id)
                        for inv_id in invocations
                    ],
                    "History URL": [
                        "{}/histories/view?id={}".format(
                            profile["galaxy_url"].strip("/"), invocations[inv_id]["history_id"]
                        )
                        for inv_id in invocations
                    ],
                },
                headers="keys",
            )
        )
    else:
        error("The tabulate package is not installed, invocations could not be listed correctly.")
        print(json.dumps(invocations, indent=4, sort_keys=True))
    info(f"{len(invocations)} invocations found.")
    return



"""Module describing :class:`planemo.interface.Engine` abstraction and implementations.

The main entrypoint is the ``contextmanager`` :func:`engine_context`.
"""

from .factory import (
    engine_context,
    is_galaxy_engine,
)

__all__ = (
    "engine_context",
    "is_galaxy_engine",
)



"""A high-level interface to local Galaxy instances using bioblend."""
from io import StringIO

from planemo.bioblend import (
    ensure_module,
    galaxy,
)

DEFAULT_ADMIN_API_KEY = "test_key"


def gi(port=None, url=None, key=None):
    """Return a bioblend ``GalaxyInstance`` for Galaxy on this port."""
    ensure_module()
    if key is None:
        key = DEFAULT_ADMIN_API_KEY
    if port is None:
        url = url
    else:
        url = "http://localhost:%d" % int(port)

    return galaxy.GalaxyInstance(url=url, key=key)


def test_credentials_valid(port=None, url=None, key=None, is_admin=False):
    """Test if provided API credentials are valid"""
    test_gi = gi(port, url, key)
    try:
        current_user = test_gi.users.get_current_user()
        if is_admin:
            return current_user["is_admin"]
        else:
            return True
    except Exception:
        return False


def user_api_key(admin_gi):
    """Use an admin authenticated account to generate a user API key."""
    ensure_module()
    # TODO: thread-safe
    users = admin_gi.users
    all_users = users.get_users()

    user_id = None
    for user in all_users:
        if user["email"] == "planemo@galaxyproject.org":
            user_id = user["id"]

    if user_id is None:
        # TODO: Allow override with --user_api_key.
        galaxy_config = admin_gi.config.get_config()
        use_remote_user = bool(galaxy_config["use_remote_user"])
        if not use_remote_user:
            user_response = users.create_local_user(
                "planemo",
                "planemo@galaxyproject.org",
                "planemo",
            )
            user_id = user_response["id"]
        else:
            user_response = users.create_remote_user(
                "planemo@galaxyproject.org",
            )
            user_id = user_response["id"]
    return users.create_user_apikey(user_id)


def summarize_history(ctx, gi, history_id):
    """Summarize a history with print() based on similar code in Galaxy for populators."""
    if not ctx.verbose:
        return

    if history_id is None:
        raise ValueError("summarize_history passed empty history_id")
    try:
        history_contents = gi.histories.show_history(history_id, contents=True)
    except Exception:
        print("Failed to fetch history contents in summarize_history.")
        return

    for history_content in history_contents:
        history_content_id = history_content.get("id", None)
        print("| %d - %s (HID - NAME) " % (int(history_content["hid"]), history_content["name"]))
        if history_content["history_content_type"] == "dataset_collection":
            history_contents_json = gi.histories.show_dataset_collection(history_id, history_content["id"])
            print("| Dataset Collection: %s" % history_contents_json)
            continue
        try:
            dataset_info = gi.histories.show_dataset(history_id, history_content_id)
            print("| Dataset State:")
            print(_format_for_summary(dataset_info.get("state"), "Dataset state is unknown."))
            print("| Dataset Blurb:")
            print(_format_for_summary(dataset_info.get("misc_blurb", ""), "Dataset blurb was empty."))
            print("| Dataset Info:")
            print(_format_for_summary(dataset_info.get("misc_info", ""), "Dataset info is empty."))
            print("| Peek:")
            print(_format_for_summary(dataset_info.get("peek", ""), "Peek unavailable."))
        except Exception:
            print("| *PLANEMO ERROR FETCHING DATASET DETAILS*")
        try:
            provenance_info = _dataset_provenance(gi, history_id, history_content_id)
            print("| Dataset Job Standard Output:")
            print(_format_for_summary(provenance_info.get("stdout", ""), "Standard output was empty."))
            print("| Dataset Job Standard Error:")
            print(_format_for_summary(provenance_info.get("stderr", ""), "Standard error was empty."))
        except Exception:
            print("| *PLANEMO ERROR FETCHING JOB DETAILS*")
        print("|")


def get_invocations(url, key, workflow_id):
    inv_gi = gi(None, url, key)
    invocations = inv_gi.workflows.get_invocations(workflow_id)
    return {
        invocation["id"]: {
            "states": inv_gi.invocations.get_invocation_summary(invocation["id"])["states"],
            "history_id": invocation["history_id"],
        }
        for invocation in invocations
    }


def _format_for_summary(blob, empty_message, prefix="|  "):
    contents = "\n".join([f"{prefix}{line.strip()}" for line in StringIO(blob).readlines() if line.rstrip("\n\r")])
    return contents or f"{prefix}*{empty_message}*"


def _dataset_provenance(gi, history_id, id):
    provenance = gi.histories.show_dataset_provenance(history_id, id)
    return provenance


__all__ = (
    "DEFAULT_ADMIN_API_KEY",
    "gi",
    "user_api_key",
)



"""Abstractions for building dependency resolution configurations."""
import tempfile
from string import Template

import click

from planemo.conda import build_conda_context

CONDA_DEPENDENCY_RESOLUTION_CONF = """<dependency_resolvers>
  <conda ${attributes} />
  <conda versionless="true" ${attributes} />
</dependency_resolvers>
"""

# Like Conda resolution above, but allow tool shed packages to be used for
# shed_serve and shed_test.
DEFAULT_DEPENDENCY_RESOLUTION_CONF = """<dependency_resolvers>
  <tool_shed_packages />
  <conda ${attributes} />
  <conda versionless="true" ${attributes} />
</dependency_resolvers>
"""

NO_DEPENDENCY_RESOLUTION_CONF = """<dependency_resolvers>
</dependency_resolvers>
"""

BREW_DEPENDENCY_RESOLUTION_CONF = """<dependency_resolvers>
  <homebrew />
  <!--
  <homebrew versionless="true" />
  -->
</dependency_resolvers>
"""

SHED_DEPENDENCY_RESOLUTION_CONF = """<dependency_resolvers>
  <tool_shed_tap />
</dependency_resolvers>
"""

# Provide some shortcuts for simple/common dependency resolutions strategies.
STOCK_DEPENDENCY_RESOLUTION_STRATEGIES = {
    "brew_dependency_resolution": BREW_DEPENDENCY_RESOLUTION_CONF,
    "shed_dependency_resolution": SHED_DEPENDENCY_RESOLUTION_CONF,
    "conda_dependency_resolution": CONDA_DEPENDENCY_RESOLUTION_CONF,
    "no_dependency_resolution": NO_DEPENDENCY_RESOLUTION_CONF,
    "default_dependency_resolution": DEFAULT_DEPENDENCY_RESOLUTION_CONF,
}


def ensure_dependency_resolvers_conf_configured(ctx, kwds, resolvers_conf=None):
    """Use supplied CLI options (kwds) to find or build a dependency resolvers file.

    Set new path in kwds if needed.
    """
    _validate_dependency_resolution_options(kwds)
    always_specify_attribute = object()

    dependency_attribute_kwds = {
        "conda_prefix": None,
        "conda_exec": None,
        "conda_debug": False,
        "conda_copy_dependencies": False,
        "conda_auto_init": always_specify_attribute,
        "conda_auto_install": always_specify_attribute,
        "conda_ensure_channels": "",
        "conda_use_local": False,
    }
    attributes = []

    def add_attribute(key, value):
        attributes.append(f'{key}="{value}"')

    conda_prefix_specified = False
    for key, default_value in dependency_attribute_kwds.items():
        value = kwds.get(key, default_value)
        if value != default_value:
            conda_prefix_specified = conda_prefix_specified or (key == "conda_prefix")
            # Strip leading prefix (conda_) off attributes
            attribute_key = "_".join(key.split("_")[1:])
            add_attribute(attribute_key, value)

    conda_context = build_conda_context(ctx, **kwds)
    if not conda_prefix_specified:
        add_attribute("prefix", conda_context.conda_prefix)
    add_attribute("condarc_override", conda_context.condarc_override)

    attribute_str = " ".join(attributes)

    if kwds.get("dependency_resolvers_config_file", None):
        resolution_type = "__explicit__"
    else:
        resolution_type = "default_dependency_resolution"
        for key in STOCK_DEPENDENCY_RESOLUTION_STRATEGIES:
            if kwds.get(key):
                resolution_type = key

    if resolution_type != "__explicit__":
        # Planemo manages the dependency resolve conf file.
        template_str = STOCK_DEPENDENCY_RESOLUTION_STRATEGIES[resolution_type]
        conf_contents = Template(template_str).safe_substitute({"attributes": attribute_str})
        if resolvers_conf is None:
            resolvers_conf = tempfile.NamedTemporaryFile(delete=False).name
        with open(resolvers_conf, "w") as fh:
            fh.write(conf_contents)
        ctx.vlog(
            "Writing dependency_resolvers_config_file to path %s with contents [%s]",
            resolvers_conf,
            conf_contents,
        )
        kwds["dependency_resolvers_config_file"] = resolvers_conf


def _validate_dependency_resolution_options(kwds):
    resolutions_strategies = [
        "brew_dependency_resolution",
        "dependency_resolvers_config_file",
        "shed_dependency_resolution",
        "conda_dependency_resolution",
    ]

    selected_strategies = 0
    for key in resolutions_strategies:
        if kwds.get(key):
            selected_strategies += 1

    if selected_strategies > 1:
        message = "At most one option from [%s] may be specified"
        raise click.UsageError(message % resolutions_strategies)



"""Module describing the planemo ``workflow_convert`` command."""
import json
import os

import click
from gxformat2 import from_galaxy_native

from planemo import options
from planemo.cli import command_function
from planemo.engine import (
    engine_context,
    is_galaxy_engine,
)
from planemo.io import write_file
from planemo.runnable import for_path


@click.command("workflow_convert")
@options.required_workflow_arg()
@options.force_option()
@options.workflow_output_artifact()
@options.galaxy_serve_options()
@command_function
def cli(ctx, workflow_identifier, output=None, force=False, **kwds):
    """Convert Format 2 workflows to native Galaxy workflows, and vice-versa."""
    assert is_galaxy_engine(**kwds)

    kwds["no_dependency_resolution"] = True

    if workflow_identifier.endswith(".ga"):
        if output is None:
            output = os.path.splitext(workflow_identifier)[0] + ".gxwf.yml"

        with open(workflow_identifier) as f:
            workflow_dict = json.load(f)
        format2_wrapper = from_galaxy_native(workflow_dict, json_wrapper=True)
        with open(output, "w") as f:
            f.write(format2_wrapper["yaml_content"])
    else:
        if output is None:
            output = os.path.splitext(workflow_identifier)[0] + ".ga"

        runnable = for_path(workflow_identifier)
        with engine_context(ctx, **kwds) as galaxy_engine:
            with galaxy_engine.ensure_runnables_served([runnable]) as config:
                workflow_id = config.workflow_id(workflow_identifier)
                output_dict = config.gi.workflows.export_workflow_dict(workflow_id)

                output_contents = json.dumps(output_dict, indent=4, sort_keys=True)
                write_file(output, output_contents, force=force)



"""Planemo specific utilities for dealing with mulled containers.

The extend galaxy-tool-util's features with planemo specific idioms.
"""

import os

from galaxy.tool_util.deps.mulled.mulled_build import (
    DEFAULT_CHANNELS,
    ensure_installed,
    InvolucroContext,
)
from galaxy.tool_util.deps.mulled.util import build_target

from planemo.conda import collect_conda_target_lists
from planemo.io import shell


def conda_to_mulled_targets(conda_targets):
    return list(map(lambda c: build_target(c.package, c.version), conda_targets))


def collect_mulled_target_lists(ctx, paths, recursive=False):
    return list(map(conda_to_mulled_targets, collect_conda_target_lists(ctx, paths, recursive=recursive)))


def build_involucro_context(ctx, **kwds):
    """Build a galaxy-tool-util InvolucroContext tailored to planemo use.

    Using planemo's common command-line/global config options.
    """
    involucro_path_default = os.path.join(ctx.workspace, "involucro")
    involucro_path = kwds.get("involucro_path", involucro_path_default)
    use_planemo_shell = kwds.get("use_planemo_shell_exec", True)
    shell_exec = shell if use_planemo_shell else None
    involucro_context = InvolucroContext(involucro_bin=involucro_path, shell_exec=shell_exec)
    if not ensure_installed(involucro_context, True):
        raise Exception("Failed to install involucro for Planemo.")
    return involucro_context


def build_mull_target_kwds(ctx, **kwds):
    """Adapt Planemo's CLI and workspace configuration to galaxy-tool-util's mulled_build options."""
    involucro_context = build_involucro_context(ctx, **kwds)
    channels = kwds.get("conda_ensure_channels", ",".join(DEFAULT_CHANNELS))
    namespace = kwds.get("mulled_namespace", "biocontainers")
    target_kwds = {
        "involucro_context": involucro_context,
        "channels": channels.split(","),
        "namespace": namespace,
        "verbose": ctx.verbose,
    }

    conda_version = kwds.get("mulled_conda_version", None)
    if conda_version is not None:
        target_kwds["conda_version"] = conda_version
    return target_kwds


__all__ = (
    "build_involucro_context",
    "build_mull_target_kwds",
    "collect_mulled_target_lists",
    "conda_to_mulled_targets",
)



"""Module describing the planemo ``docs`` command."""
import click

from planemo.cli import command_function

DOCS_URL = "http://planemo.readthedocs.io/en/latest/"


@click.command("docs")
@command_function
def cli(ctx, **kwds):
    """Open Planemo documentation in web browser."""
    click.launch(DOCS_URL)



"""Docker utilities for planemo.

Built on Galaxy abstractions in :mod:`galaxy.tools.deps.dockerfiles` and
:mod:`galaxy.tools.deps.docker_util`.
"""

from galaxy.tool_util.deps.dockerfiles import docker_host_args

__all__ = ("docker_host_args",)



"""Module describing the planemo ``config_init`` command."""
import os
import sys

import click

from planemo import (
    config,
    options,
)
from planemo.cli import command_function
from planemo.io import (
    info,
    warn,
)

CONFIG_TEMPLATE = """## Planemo Global Configuration File.
## Everything in this file is completely optional - these values can all be
## configured via command line options for the corresponding commands.

## Specify a default galaxy_root for test and server commands here.
#galaxy_root: /path/to/galaxy_root
## Username used with toolshed(s).
#shed_username: "<TODO>"
sheds:
  # For each tool shed you wish to target, uncomment key or both email and
  # password.
  toolshed:
    #key: "<TODO>"
    #email: "<TODO>"
    #password: "<TODO>"
  testtoolshed:
    #key: "<TODO>"
    #email: "<TODO>"
    #password: "<TODO>"
  local:
    #key: "<TODO>"
    #email: "<TODO>"
    #password: "<TODO>"
"""
SUCCESS_MESSAGE = "Wrote configuration template to %s, please open with editor and fill out."


@click.command("config_init")
@options.optional_project_arg(exists=None)
@click.option("--template", default=None)
@command_function
def cli(ctx, path, template=None, **kwds):
    """Initialise global configuration for Planemo.

    Helps initialize global configuration (in home directory) for Planemo.
    """
    # TODO: prompt for values someday.
    config_path = config.global_config_path()
    if os.path.exists(config_path):
        warn("File %s already exists, exiting." % config_path)
        sys.exit(1)
    with open(config_path, "w") as f:
        f.write(CONFIG_TEMPLATE)
        info(SUCCESS_MESSAGE % config_path)



"""Module describing the planemo ``shed_build`` command."""

import shutil
import sys

import click

from planemo import (
    options,
    shed,
)
from planemo.cli import command_function


@click.command("shed_build")
@options.optional_tools_arg(multiple=False)
@command_function
def cli(ctx, path, **kwds):
    """Create a Galaxy tool tarball.

    This will use the .shed.yml file to prepare a tarball
    (which you could upload to the Tool Shed manually).
    """

    def build(realized_repository):
        tarpath = shed.build_tarball(realized_repository.path)
        outpath = realized_repository.real_path + ".tar.gz"
        shutil.move(tarpath, outpath)
        print("Created: %s" % (outpath))
        return 0

    exit_code = shed.for_each_repository(ctx, build, [path], **kwds)
    sys.exit(exit_code)



"""Utilities for calculating effective repository diffs.

Some intelligence is required because the tool shed updates attributes that it
is beneficial to ignore.
"""

import os
import sys
from xml.etree import ElementTree

from planemo.xml import diff


def diff_and_remove(working, label_a, label_b, f):
    """Remove tool shed XML files and use a smart XML diff on them.

    Return 0 if and only if the XML content is the sam after stripping
    attirbutes the tool shed updates.
    """
    assert label_a != label_b
    special = ["tool_dependencies.xml", "repository_dependencies.xml"]
    deps_diff = 0
    # Could walk either A or B; will only compare if in same relative location
    for dirpath, dirnames, filenames in os.walk(os.path.join(working, label_a)):
        for filename in filenames:
            if filename in special:
                a = os.path.join(dirpath, filename)
                b = os.path.join(working, label_b, os.path.relpath(a, os.path.join(working, label_a)))
                files_exist = os.path.exists(a) and os.path.exists(b)
                if files_exist:
                    deps_diff |= _shed_diff(a, b, f)
                    os.remove(a)
                    os.remove(b)
    return deps_diff


def _shed_diff(file_a, file_b, f=sys.stdout):
    """Strip attributes the tool shed writes and do smart XML diff.

    Returns 0 if and only if the XML content is the same after stripping
    ``tool_shed`` and ``changeset_revision`` attributes.
    """
    xml_a = ElementTree.parse(file_a).getroot()
    xml_b = ElementTree.parse(file_b).getroot()
    _strip_shed_attributes(xml_a)
    _strip_shed_attributes(xml_b)
    return diff.diff(xml_a, xml_b, reporter=f.write)


def _strip_shed_attributes(xml_element):
    if xml_element.tag == "repository":
        _remove_attribs(xml_element)
    for child in xml_element:
        _strip_shed_attributes(child)


def _remove_attribs(xml_element):
    for attrib in ["changeset_revision", "toolshed"]:
        if attrib in xml_element.attrib:
            del xml_element.attrib[attrib]


__all__ = ("diff_and_remove",)



"""Module describing the planemo ``shed_upload`` command."""
import sys

import click

from planemo import (
    options,
    shed,
)
from planemo.cli import command_function

tar_path = click.Path(
    exists=True,
    file_okay=True,
    dir_okay=False,
    resolve_path=True,
)


@click.command("shed_upload")
@options.shed_publish_options()
@options.shed_upload_options()
@click.option(
    "--tar_only",
    is_flag=True,
    help="Produce tar file for upload but do not publish to a tool shed.",
)
@click.option(
    "--tar",
    help="Specify a pre-existing tar file instead of automatically building " "one as part of this command.",
    type=tar_path,
    default=None,
)
@command_function
def cli(ctx, paths, **kwds):
    """Low-level command to upload tarballs.

    Generally, ``shed_update`` should be used instead since it also updates
    both tool shed contents (via tar ball generation and upload) as well as
    metadata (to handle metadata changes in ``.shed.yml`` files).

    \b
        % planemo shed_upload --tar_only  ~/
        % tar -tzf shed_upload.tar.gz
        test-data/blastdb.loc
        ...
        tools/ncbi_blast_plus/tool_dependencies.xml
        % tar -tzf shed_upload.tar.gz | wc -l
        117

    """

    def upload(realized_repository):
        return shed.upload_repository(ctx, realized_repository, **kwds)

    exit_code = shed.for_each_repository(ctx, upload, paths, **kwds)
    sys.exit(exit_code)



"""Check an output file from a generalize artifact test."""

import os
import tempfile

import requests
from galaxy.tool_util.parser.interface import TestCollectionOutputDef
from galaxy.tool_util.verify import verify
from galaxy.tool_util.verify.interactor import verify_collection
from galaxy.util import unicodify


def check_output(runnable, output_properties, test_properties, **kwds):
    """Use galaxy-tool-util to check a test output.

    Return a list of strings describing the problems encountered,
    and empty list indicates no problems were detected.

    Currently this will only ever return at most one detected problem because
    of the way galaxy-tool-util throws exceptions instead of returning individual
    descriptions - but this may be enhanced in the future.
    """
    checker = _check_output_collection if for_collections(test_properties) else _check_output_file
    return checker(runnable, output_properties, test_properties, **kwds)


def for_collections(test_properties):
    return "element_tests" in test_properties


def _check_output_collection(runnable, output_properties, test_properties, **kwds):
    data_collection = output_properties

    output_def = TestCollectionOutputDef.from_dict(test_properties)

    def verify_dataset(element, element_attrib, element_outfile):
        if element_outfile:
            element_attrib["path"] = element_outfile
        _verify_output_file(runnable, element["_output_object"], element_attrib)

    problems = []
    try:
        verify_collection(output_def, data_collection, verify_dataset)
    except AssertionError as e:
        problems.append(unicodify(e))

    return problems


def _verify_output_file(runnable, output_properties, test_properties, **kwds):
    get_filename = _test_filename_getter(runnable)
    path = output_properties["path"]
    with open(path, "rb") as fh:
        output_content = fh.read()
    # Support Galaxy-like file location (using "file") or CWL-like ("path" or "location").
    expected_file = test_properties.get("file", None)
    if expected_file is None:
        expected_file = test_properties.get("path", None)
    if expected_file is None:
        location = test_properties.get("location")
        if location:
            if location.startswith(("http://", "https://")):
                expected_file = _get_location(location)
            else:
                expected_file = location.split("file://", 1)[-1]

    test_data_target_dir = kwds.get("test_data_target_dir", None)
    item_label = "Output with path %s" % path
    if "asserts" in test_properties:
        # TODO: break fewer abstractions here...
        from galaxy.tool_util.parser.yaml import __to_test_assert_list

        test_properties["assert_list"] = __to_test_assert_list(test_properties["asserts"])
    verify(
        item_label,
        output_content,
        attributes=test_properties,
        filename=expected_file,
        get_filename=get_filename,
        keep_outputs_dir=test_data_target_dir,
        verify_extra_files=None,
    )


def _check_output_file(runnable, output_properties, test_properties, **kwds):
    problems = []
    try:
        _verify_output_file(runnable, output_properties, test_properties, **kwds)
    except AssertionError as e:
        problems.append(unicodify(e))

    return problems


def _get_location(location):
    data_file = tempfile.NamedTemporaryFile(prefix="planemo_test_file_", delete=False)
    with requests.get(location, stream=True) as r:
        r.raise_for_status()

        for chunk in r.iter_content():
            if chunk:
                data_file.write(chunk)
        return data_file.name


def _test_filename_getter(runnable):
    def get_filename(name):
        artifact_directory = os.path.dirname(runnable.path)
        return os.path.join(artifact_directory, name)

    return get_filename


__all__ = ("check_output",)



"""Module contianing the :class:`Engine` abstraction."""

import abc
import json
import os
import tempfile
from typing import List

from planemo.exit_codes import EXIT_CODE_UNSUPPORTED_FILE_TYPE
from planemo.io import error
from planemo.runnable import (
    cases,
    RunnableType,
)
from planemo.test.results import StructuredData


class Engine(metaclass=abc.ABCMeta):
    """Abstract description of an external process for running tools or workflows."""

    @abc.abstractmethod
    def run(self, path, job_path):
        """Run a job using a compatible artifact (workflow or tool)."""

    @abc.abstractmethod
    def cleanup(self):
        """Release any resources used to run/test with this engine."""

    @abc.abstractmethod
    def test(self, runnables):
        """Test runnable artifacts (workflow or tool)."""


class BaseEngine(Engine):
    """Base class providing context and keywords for Engine implementations."""

    handled_runnable_types = []  # type: List[RunnableType]

    def __init__(self, ctx, **kwds):
        """Store context and kwds."""
        self._ctx = ctx
        self._kwds = kwds

    def can_run(self, runnable):
        """Use subclass's ``handled_runnable_types`` variable to infer ``can_run``."""
        return runnable.type in self.handled_runnable_types

    def cleanup(self):
        """Default no-op cleanup method."""

    def run(self, runnables, job_paths):
        """Run a job using a compatible artifact (workflow or tool)."""
        self._check_can_run_all(runnables)
        run_responses = self._run(runnables, job_paths)
        return run_responses

    @abc.abstractmethod
    def _run(self, runnables, job_path):
        """Run a job using a compatible artifact (workflow or tool) wrapped as a runnable."""

    def _check_can_run(self, runnable):
        if not self.can_run(runnable):
            template = "Engine type [%s] cannot execute [%s]s"
            message = template % (self.__class__, runnable.type)
            error(message)
            self._ctx.exit(EXIT_CODE_UNSUPPORTED_FILE_TYPE)

    def _check_can_run_all(self, runnables):
        for runnable in runnables:
            self._check_can_run(runnable)

    def test(self, runnables, test_timeout):
        """Test runnable artifacts (workflow or tool)."""
        self._check_can_run_all(runnables)
        test_cases = [t for tl in map(cases, runnables) for t in tl]
        test_results = self._collect_test_results(test_cases, test_timeout)
        tests = []
        for (test_case, run_response) in test_results:
            test_case_data = test_case.structured_test_data(run_response)
            tests.append(test_case_data)
        test_data = {
            "version": "0.1",
            "tests": tests,
        }
        structured_results = StructuredData(data=test_data)
        structured_results.calculate_summary_data()
        return structured_results

    def _collect_test_results(self, test_cases, test_timeout):
        run_responses = self._run_test_cases(test_cases, test_timeout)
        return [(test_case, run_response) for test_case, run_response in zip(test_cases, run_responses)]

    def _run_test_cases(self, test_cases, test_timeout):
        runnables = [test_case.runnable for test_case in test_cases]
        job_paths = []
        tmp_paths = []
        for test_case in test_cases:
            if test_case.job_path is None:
                job = test_case.job
                with tempfile.NamedTemporaryFile(
                    dir=test_case.tests_directory,
                    suffix=".json",
                    prefix="plnmotmptestjob",
                    delete=False,
                    mode="w+",
                ) as f:
                    tmp_path = f.name
                    job_path = tmp_path
                    tmp_paths.append(tmp_path)
                    json.dump(job, f)
                job_paths.append(job_path)
            else:
                job_paths.append(test_case.job_path)
        try:
            run_responses = self._run(runnables, job_paths)
        finally:
            for tmp_path in tmp_paths:
                os.remove(tmp_path)
        return run_responses


__all__ = (
    "Engine",
    "BaseEngine",
)



"""Utilities for reasoning about Galaxy test results."""

import os
import shlex
from typing import NamedTuple
from xml.etree import ElementTree as ET

from planemo.io import error
from planemo.test.results import StructuredData as BaseStructuredData

NO_STRUCTURED_FILE = (
    "Warning: Problem with target Galaxy, it did not "
    "produce a structured test results file [%s] - summary "
    "information and planemo reports will be incorrect."
)


class GalaxyTestCommand:
    """Abstraction around building a ``run_tests.sh`` command for Galaxy tests."""

    def __init__(
        self,
        html_report_file,
        xunit_report_file,
        structured_report_file,
        failed=False,
        installed=False,
    ):
        self.html_report_file = html_report_file
        self.xunit_report_file = xunit_report_file
        self.structured_report_file = structured_report_file
        self.failed = failed
        self.installed = installed

    def build(self):
        xunit_report_file = self.xunit_report_file
        sd_report_file = self.structured_report_file
        cmd = "./run_tests.sh $COMMON_STARTUP_ARGS --report_file %s" % shlex.quote(self.html_report_file)
        if xunit_report_file:
            cmd += " --xunit_report_file %s" % shlex.quote(xunit_report_file)
        if sd_report_file:
            cmd += " --structured_data_report_file %s" % shlex.quote(sd_report_file)
        if self.installed:
            cmd += " -installed"
        elif self.failed:
            sd = StructuredData(sd_report_file)
            tests = " ".join(sd.failed_ids)
            cmd += " %s" % tests
        else:
            cmd += " functional.test_toolbox"
        return cmd


class StructuredData(BaseStructuredData):
    """Abstraction around Galaxy's structured test data output."""

    def __init__(self, json_path):
        if not json_path or not os.path.exists(json_path):
            error(NO_STRUCTURED_FILE % json_path)
        super().__init__(json_path)

    def merge_xunit(self, xunit_root):
        self.has_details = True
        xunit_attrib = xunit_root.attrib
        num_tests = int(xunit_attrib.get("tests", 0))
        num_failures = int(xunit_attrib.get("failures", 0))
        num_errors = int(xunit_attrib.get("errors", 0))
        num_skips = int(xunit_attrib.get("skips", 0))
        summary = dict(
            num_tests=num_tests,
            num_failures=num_failures,
            num_errors=num_errors,
            num_skips=num_skips,
        )

        self.structured_data["summary"] = summary

        for testcase_el in xunit_t_elements_from_root(xunit_root):
            test = case_id(testcase_el)
            test_data = self.structured_data_by_id.get(test.id)
            if not test_data:
                continue
            problem_el = None
            for problem_type in ["skip", "failure", "error"]:
                problem_el = testcase_el.find(problem_type)
                if problem_el is not None:
                    break
            if problem_el is not None:
                status = problem_el.tag
                test_data["problem_type"] = problem_el.attrib["type"]
                test_data["problem_log"] = problem_el.text
            else:
                status = "success"
            test_data["status"] = status


class GalaxyTestResults:
    """Class that combine the test-centric xunit output
    with the Galaxy centric structured data output - and
    abstracts away the difference (someday).
    """

    def __init__(
        self,
        output_json_path,
        output_xml_path,
        output_html_path,
        exit_code,
    ):
        self.output_html_path = output_html_path
        sd = StructuredData(output_json_path)
        self.sd = sd
        self.structured_data = sd.structured_data
        self.structured_data_tests = sd.structured_data_tests
        self.structured_data_by_id = sd.structured_data_by_id

        self.xunit_tree = parse_xunit_report(output_xml_path)
        sd.merge_xunit(self._xunit_root)

        self.sd.set_exit_code(exit_code)
        self.sd.read_summary()
        self.sd.update()

    @property
    def exit_code(self):
        return self.sd.exit_code

    @property
    def has_details(self):
        return self.sd.has_details

    @property
    def num_tests(self):
        return self.sd.num_tests

    @property
    def num_problems(self):
        return self.sd.num_problems

    @property
    def _xunit_root(self):
        return self.xunit_tree.getroot()

    @property
    def all_tests_passed(self):
        return self.sd.num_problems == 0

    @property
    def xunit_testcase_elements(self):
        return xunit_t_elements_from_root(self._xunit_root)


def xunit_t_elements_from_root(xunit_root):
    yield from find_cases(xunit_root)


def parse_xunit_report(xunit_report_path):
    return ET.parse(xunit_report_path)


def find_cases(xunit_root):
    return xunit_root.findall("testcase")


def case_id(testcase_el=None, raw_id=None):
    if raw_id is None:
        assert testcase_el is not None
        name_raw = testcase_el.attrib["name"]
        if "TestForTool_" in name_raw:
            raw_id = name_raw
        else:
            class_name = testcase_el.attrib["classname"]
            raw_id = f"{class_name}.{name_raw}"

    name = None
    num = None
    if "TestForTool_" in raw_id:
        tool_and_num = raw_id.split("TestForTool_")[-1]
        if ".test_tool_" in tool_and_num:
            name, num_str = tool_and_num.split(".test_tool_", 1)
            num = _parse_num(num_str)
            # Tempted to but something human friendly in here like
            # num + 1 - but then it doesn't match HTML report.
        else:
            name = tool_and_num
    else:
        name = raw_id

    return TestId(name, num, raw_id)


def _parse_num(num_str):
    try:
        num = int(num_str)
    except ValueError:
        num = None
    return num


class TestId(NamedTuple):
    name: str
    num: int
    id: str

    @property
    def label(self):
        if self.num is not None:
            return f"{self.name}[{self.num}]"
        else:
            return self.id



"""Module describes a :class:`DatabaseSource` for managed, dockerized postgres databases."""
import time

from galaxy.tool_util.deps import (
    docker_util,
    dockerfiles,
)
from galaxy.util import unicodify
from galaxy.util.commands import execute

from .interface import DatabaseSource
from .postgres import (
    _CommandBuilder,
    ExecutesPostgresSqlMixin,
)

DEFAULT_CONTAINER_NAME = "planemopostgres"
DEFAULT_POSTGRES_PASSWORD = "mysecretpassword"
DEFAULT_POSTGRES_PORT_EXPOSE = 15432


def docker_ps(args, **kwds):
    return docker_util.command_list("ps", args, **kwds)


def docker_exec(name, commands=[], **kwds):
    return docker_util.command_list("exec", [name] + commands, **kwds)


def is_running_container(name=DEFAULT_CONTAINER_NAME, **kwds):
    ps_command = docker_ps(["--format", "{{.Names}}"], **kwds)
    running_containers = unicodify(execute(ps_command))
    containers = running_containers.splitlines()
    return name in containers


def start_postgres_docker(
    name=DEFAULT_CONTAINER_NAME, password=DEFAULT_POSTGRES_PASSWORD, port=DEFAULT_POSTGRES_PORT_EXPOSE, **kwds
):
    run_command = docker_util.command_list(
        "run",
        ["-p", "%d:5432" % port, "--name", name, "-e", "POSTGRES_PASSWORD=%s" % password, "--rm", "-d", "postgres"],
        **kwds,
    )
    execute(run_command)


def stop_postgres_docker(name=DEFAULT_CONTAINER_NAME, **kwds):
    stop_command = docker_util.command_list("stop", [name], **kwds)
    execute(stop_command)


class DockerPostgresDatabaseSource(ExecutesPostgresSqlMixin, DatabaseSource):
    """Postgres database running inside a Docker container."""

    def __init__(self, **kwds):
        """Construct a postgres database source from planemo configuration."""
        self.psql_path = "psql"
        self.database_user = "postgres"
        self.database_password = DEFAULT_POSTGRES_PASSWORD
        self.database_host = "localhost"  # TODO: Make docker host
        self.database_port = DEFAULT_POSTGRES_PORT_EXPOSE
        self._kwds = kwds
        self._docker_host_kwds = dockerfiles.docker_host_args(**kwds)
        if not is_running_container(**self._docker_host_kwds):
            start_postgres_docker(**self._docker_host_kwds)
            # Hack to give docker a bit of time to boot up and allow psql to start.
            time.sleep(30)

    def sqlalchemy_url(self, identifier):
        """Return URL or form postgresql://username:password@localhost/mydatabase."""
        return "postgresql://%s:%s@%s:%d/%s" % (
            self.database_user,
            self.database_password,
            self.database_host,
            self.database_port,
            identifier,
        )

    def _psql_command_builder(self, *args):
        base_command = docker_exec(DEFAULT_CONTAINER_NAME, [self.psql_path], **self._docker_host_kwds)
        command_builder = _CommandBuilder(*base_command)
        # Print only tuples so output is easier to parse
        command_builder.append_command("--tuples-only")
        command_builder.append_command("--username", self.database_user)
        command_builder.append_command("-P", "pager=off")
        command_builder.extend_command(args)
        return command_builder


__all__ = ("DockerPostgresDatabaseSource",)



"""Module describing the planemo ``list_repos`` command."""

import click

from planemo import options
from planemo.ci import print_as_yaml
from planemo.cli import command_function
from planemo.shed import (
    _realize_effective_repositories,
    find_raw_repositories,
)


@click.command("list_repos")
@options.shed_project_arg()
@options.ci_find_options()
@command_function
def cli(ctx, paths, **kwds):
    """Find all shed repositories in one or more directories and output as yaml.

    Currently, a shed repository is considered a directory with a .shed.yml
    file.
    """
    kwds["recursive"] = True
    kwds["fail_fast"] = True
    repos = find_raw_repositories(ctx, paths, **kwds)
    # Since fail_fast is True, all repos are actual raw repo objects and
    # not exceptions.
    raw_paths = [r.path for r in repos]
    realized_repos = []
    for repo_gen in (_realize_effective_repositories(r, path=p) for r, p in zip(repos, raw_paths)):
        for repo in repo_gen:
            realized_repos.append({"name": repo.config["name"], "owner": repo.config["owner"]})
    print_as_yaml(realized_repos, **kwds)



"""Utilities for calling Galaxy scripts."""
import os
import shlex
import string

from galaxy.util.commands import shell

from planemo.io import (
    info,
    shell_join,
)
from planemo.virtualenv import (
    create_command,
    DEFAULT_PYTHON_VERSION,
)

# Activate galaxy's virtualenv if present (needed for tests say but not for
# server because run.sh does this).
ACTIVATE_COMMAND = (
    'if [ -e "$GALAXY_VIRTUAL_ENV" ]; then . "$GALAXY_VIRTUAL_ENV"/bin/activate; '
    'echo "Activated a virtualenv for Galaxy"; echo "$VIRTUAL_ENV"; '
    'else echo "Failed to activate virtualenv."; fi'
)
CREATE_COMMAND_TEMPLATE = string.Template(
    'if [ ! -e "$GALAXY_VIRTUAL_ENV" ]; then $create_virtualenv; echo "Created virtualenv"; fi',
)
PRINT_VENV_COMMAND = shell_join(
    r'echo "Set \$GALAXY_VIRTUAL_ENV to $GALAXY_VIRTUAL_ENV"',
    (
        'if [ -e "$GALAXY_VIRTUAL_ENV" ]; '
        'then echo "Virtual environment directory exists."; '
        'else echo "Virtual environment directory does not exist."; fi'
    ),
)


CACHED_VIRTUAL_ENV_COMMAND = "if [ -d .venv ]; " "then GALAXY_VIRTUAL_ENV=.venv; " "else GALAXY_VIRTUAL_ENV=%s; fi"
UNCACHED_VIRTUAL_ENV_COMMAND = "GALAXY_VIRTUAL_ENV=.venv"


def setup_venv(ctx, kwds):
    if kwds.get("skip_venv", False):
        return ""

    create_template_params = {
        "create_virtualenv": create_command("$GALAXY_VIRTUAL_ENV", kwds.get("galaxy_python_version"))
    }
    return shell_join(
        locate_galaxy_virtualenv(ctx, kwds),
        PRINT_VENV_COMMAND if ctx.verbose else None,
        CREATE_COMMAND_TEMPLATE.safe_substitute(create_template_params),
        PRINT_VENV_COMMAND if ctx.verbose else None,
        ACTIVATE_COMMAND,
        "bash -c 'pwd'" if ctx.verbose else None,
        "bash -c 'which python'" if ctx.verbose else None,
        "bash -c 'which pip'" if ctx.verbose else None,
        "bash -c 'echo $GALAXY_VIRTUAL_ENV'" if ctx.verbose else None,
        "bash -c 'echo $VIRTUAL_ENV'" if ctx.verbose else None,
        "bash -c 'ls -a'" if ctx.verbose else None,
    )


def locate_galaxy_virtualenv(ctx, kwds):
    if os.environ.get("GALAXY_VIRTUAL_ENV"):
        venv_command = ""
        kwds["GALAXY_VIRTUAL_ENV"] = os.environ["GALAXY_VIRTUAL_ENV"]
    elif not kwds.get("no_cache_galaxy", False):
        workspace = ctx.workspace
        galaxy_branch = kwds.get("galaxy_branch") or "master"
        shared_venv_path = os.path.join(workspace, "gx_venv")
        galaxy_python_version = kwds.get("galaxy_python_version") or DEFAULT_PYTHON_VERSION
        shared_venv_path = f"{shared_venv_path}_{galaxy_python_version}"
        if galaxy_branch != "master":
            shared_venv_path = f"{shared_venv_path}_{galaxy_branch}"
        kwds["GALAXY_VIRTUAL_ENV"] = shared_venv_path
        venv_command = CACHED_VIRTUAL_ENV_COMMAND % shlex.quote(shared_venv_path)
    else:
        kwds["GALAXY_VIRTUAL_ENV"] = ".venv"
        venv_command = UNCACHED_VIRTUAL_ENV_COMMAND
    return shell_join(
        venv_command,
        "export GALAXY_VIRTUAL_ENV",
    )


def run_galaxy_command(ctx, command, env, action):
    """Run Galaxy command with informative verbose logging."""
    message = f"{action} with command [{command}]"
    # info not working in pytest+Github actions the way it did in nose?
    info(message)
    ctx.vlog("With environment variables:")
    ctx.vlog("============================")
    for key, value in env.items():
        ctx.vlog(f'{key}="{value}"')
    ctx.vlog("============================")
    exit_code = shell(command, env=env)
    ctx.vlog("run command exited with return code %s" % exit_code)
    return exit_code


__all__ = (
    "setup_venv",
    "run_galaxy_command",
)



"""Module describing the planemo ``conda_search`` command."""

import click
import packaging.version

from planemo import options
from planemo.cli import command_function
from planemo.conda import build_conda_context

VERSION_4_DOT_4 = packaging.version.Version("4.4")


@click.command("conda_search")
@options.conda_target_options(include_local=False)
@click.argument(
    "term",
    metavar="TERM",
    type=str,
    nargs=1,
)
@command_function
def cli(ctx, term, **kwds):
    """Perform conda search with Planemo's conda.

    Implicitly adds channels Planemo is configured with.
    """
    conda_context = build_conda_context(ctx, handle_auto_init=True, **kwds)
    # Handle CLI interface change for conda search in 4.5.
    #  xref: https://github.com/conda/conda/pull/5597/files
    if conda_context.conda_version >= VERSION_4_DOT_4:
        term = "*%s*" % term
    args = conda_context._override_channels_args + [term]
    conda_context.exec_command("search", args)



def diff(x1, x2, reporter=None):
    """Return 0 if and only if the XML has the same content."""
    compare = xml_compare(x1, x2, reporter)
    return_val = 0 if compare else 1
    return return_val


# From
# bitbucket.org/ianb/formencode/src/tip/formencode/doctest_xml_compare.py
# with (PSF license)
def xml_compare(x1, x2, reporter=None):
    if reporter is None:

        def reporter(x):
            return None

    if x1.tag != x2.tag:
        reporter(f"Tags do not match: {x1.tag} and {x2.tag}\n")
        return False
    for name, value in x1.attrib.items():
        if x2.attrib.get(name) != value:
            reporter("Attributes do not match: %s=%r, %s=%r\n" % (name, value, name, x2.attrib.get(name)))
            return False
    for name in x2.attrib.keys():
        if name not in x1.attrib:
            reporter("x2 has an attribute x1 is missing: %s\n" % name)
            return False
    if not text_compare(x1.text, x2.text):
        reporter(f"text: {x1.text!r} != {x2.text!r}\n")
        return False
    if not text_compare(x1.tail, x2.tail):
        reporter(f"tail: {x1.tail!r} != {x2.tail!r}\n")
        return False
    return _compare_children(x1, x2, reporter)


def _compare_children(x1, x2, reporter):
    cl1 = list(x1)
    cl2 = list(x2)
    if len(cl1) != len(cl2):
        reporter("children length differs, %i != %i\n" % (len(cl1), len(cl2)))
        return False
    i = 0
    for c1, c2 in zip(cl1, cl2):
        i += 1
        if not xml_compare(c1, c2, reporter=reporter):
            reporter("children %i do not match: %s\n" % (i, c1.tag))
            return False
    return True


def text_compare(t1, t2):
    if not t1 and not t2:
        return True
    return (t1 or "").strip() == (t2 or "").strip()



"""Describes results.

Is a JSON.
"""
import json
import os

from planemo.io import error


class StructuredData:
    """Abstraction around a simple data structure describing test results."""

    def __init__(self, json_path=None, data=None):
        """Create a :class:`StructuredData` from a JSON file."""

        def data_error():
            error(
                "An invalid JSON for structured test result data - "
                "summary information and planemo reports will be "
                "incorrect."
            )

        self.json_path = json_path
        structured_data = {}
        structured_data_tests = {}
        if json_path and os.path.exists(json_path) and data is None:
            try:
                with open(json_path) as output_json_f:
                    data = json.load(output_json_f)
            except Exception:
                data_error()

        try:
            structured_data = data
            structured_data_tests = structured_data["tests"]
        except Exception:
            data_error()

        self.structured_data = structured_data
        self.structured_data_tests = structured_data_tests
        structured_data_by_id = {}
        for test in self.structured_data_tests:
            structured_data_by_id[test["id"]] = test["data"]
        self.structured_data_by_id = structured_data_by_id
        self.has_details = "summary" in structured_data
        if self.has_details:
            self.read_summary()

    def update(self):
        """Write out an updated version of this data structure to supplied json path."""
        with open(self.json_path, "w") as out_f:
            json.dump(self.structured_data, out_f)

    def set_exit_code(self, exit_code):
        """Set the exit_code for the this test."""
        self.structured_data["exit_code"] = exit_code

    def calculate_summary_data_if_needed(self):
        if "summary" not in self.structured_data:
            self.calculate_summary_data()

    def calculate_summary_data(self):
        """Use full details on individual test data to update structured data with summary info."""
        num_tests = 0
        num_failures = 0
        num_skips = 0
        num_errors = 0

        for test in self.structured_data_tests:
            test_data = get_dict_value("data", test)
            status = get_dict_value("status", test_data)
            num_tests += 1
            if status == "skip":
                num_skips += 1
            elif status == "failure":
                num_failures += 1
            elif status == "error":
                num_errors += 1
            elif status != "success":
                raise Exception("Unknown test status encountered [%s]" % status)

        summary = {}
        summary["num_tests"] = num_tests
        summary["num_failures"] = num_failures
        summary["num_skips"] = num_skips
        summary["num_errors"] = num_errors
        self.structured_data["summary"] = summary
        self.read_summary()

    def read_summary(self):
        """Read summary data into properties on this class."""
        summary = self.structured_data["summary"]
        num_tests = summary["num_tests"]
        num_failures = summary["num_failures"]
        num_skips = summary["num_skips"]
        num_errors = summary["num_errors"]

        self.num_tests = num_tests
        self.num_problems = num_skips + num_errors + num_failures

        self.exit_code = self.structured_data.get("exit_code", None)

    @property
    def failed_ids(self):
        """Find set of IDs for failed tests."""
        ids = set()
        for test_data in self.structured_data_tests:
            if test_data["data"]["status"] == "success":
                continue
            test_case = test_data["id"].replace(".test_toolbox.", ".test_toolbox:")
            ids.add(test_case)
        return ids


def get_dict_value(key, data):
    """Return data[key] with improved KeyError."""
    try:
        return data[key]
    except (KeyError, TypeError):
        raise KeyError(f"No key [{key}] in [{data}]")


__all__ = (
    "StructuredData",
    "get_dict_value",
)



"""Utilities for dealing with continous integration systems."""


import copy
import math
import os

import yaml

from planemo import (
    git,
    io,
)
from planemo.shed import REPO_METADATA_FILES


def filter_paths(ctx, raw_paths, path_type="repo", **kwds):
    """Filter ``paths``.

    ``path_type`` is ``repo`` or ``file``.
    """
    cwd = os.getcwd()

    filter_kwds = copy.deepcopy(kwds)
    changed_in_commit_range = kwds.get("changed_in_commit_range", None)
    diff_paths = None
    if changed_in_commit_range is not None:
        diff_files = git.diff(ctx, cwd, changed_in_commit_range)
        if path_type == "repo":
            diff_dirs = {os.path.dirname(p) for p in diff_files}
            diff_paths = set()
            for diff_dir in diff_dirs:
                diff_path = metadata_file_in_path(diff_dir)
                if diff_path:
                    diff_paths.add(diff_path)
        else:
            diff_paths = diff_files

    unique_paths = {os.path.relpath(p, cwd) for p in raw_paths}
    if diff_paths is not None:
        unique_paths = unique_paths.intersection(diff_paths)
    filtered_paths = sorted(io.filter_paths(unique_paths, cwd=cwd, **filter_kwds))
    excluded_paths = sorted(set(unique_paths) - set(filtered_paths))
    if excluded_paths:
        ctx.log("List of excluded paths: %s" % excluded_paths)

    path_count = len(filtered_paths)
    chunk_size = (1.0 * path_count) / kwds["chunk_count"]
    chunk = kwds["chunk"]

    chunked_paths = []
    for i, path in enumerate(filtered_paths):
        if int(math.floor(i / chunk_size)) == chunk:
            chunked_paths.append(path)

    return chunked_paths


def metadata_file_in_path(diff_dir):
    while diff_dir:
        for metadata_file in REPO_METADATA_FILES:
            if os.path.isfile(os.path.join(diff_dir, metadata_file)):
                return diff_dir
        diff_dir = os.path.dirname(diff_dir)


def group_paths(paths):
    repos = {}
    for path in paths:
        repo = os.path.split(path)[0]
        if repo not in repos:
            repos[repo] = []
        repos[repo].append(path)
    return [" ".join(repos[_]) for _ in repos]


def print_path_list(paths, **kwds):
    with io.open_file_or_standard_output(kwds["output"], "w") as f:
        for path in paths:
            print(path, file=f)


def print_as_yaml(item, **kwds):
    with io.open_file_or_standard_output(kwds["output"], "w") as f:
        f.write(yaml.safe_dump(item))
